# flake8:noqa:F401
from nonebot.adapters.cqhttp.message import Message
from nonebot.adapters.cqhttp.utils import escape, unescape
from nonebot.exception import ActionFailed, ApiNotAvailable

from .bus import EventBus
from .comp import NoneBot
from .exceptions import NoneTripCompException as CQHttpError
from .poly import CQHttp, Event



import asyncio
import warnings
from typing import Any, Iterable, List, NamedTuple, Optional, Set, Tuple, Union

from nonebot.adapters.cqhttp.message import Message

from ..poly import Event as CQEvent
from . import NoneBot
from .command import call_command
from .log import logger
from .session import BaseSession
from .typing import CommandArgs_T, CommandName_T, NLPHandler_T, PermChecker_T


class NLProcessor:
    """INTERNAL API"""

    __slots__ = (
        "func",
        "keywords",
        "only_to_me",
        "only_short_message",
        "allow_empty_message",
        "perm_checker_func",
    )

    def __init__(
        self,
        *,
        func: NLPHandler_T,
        keywords: Optional[Iterable[str]],
        only_to_me: bool,
        only_short_message: bool,
        allow_empty_message: bool,
        perm_checker_func: PermChecker_T,
    ):
        self.func = func
        self.keywords = keywords
        self.only_to_me = only_to_me
        self.only_short_message = only_short_message
        self.allow_empty_message = allow_empty_message
        self.perm_checker_func = perm_checker_func  # returns True if can trigger

    async def test(
        self, session: "NLPSession", msg_text_length: Optional[int] = None
    ) -> bool:
        """
        Test whether the session matches this (self) NL processor.

        :param session: NLPSession object
        :param msg_text_length: this argument is `len(session.msg_text)`,
                                designated to be cached if this function
                                is invoked in a loop
        :return: the session context matches this processor
        """
        if msg_text_length is None:
            msg_text_length = len(session.msg_text)

        if not self.allow_empty_message and not session.msg:
            # don't allow empty msg, but it is one, so no
            return False

        if (
            self.only_short_message
            and msg_text_length > session.bot.config.SHORT_MESSAGE_MAX_LENGTH
        ):
            return False

        if self.only_to_me and not session.event["to_me"]:
            return False

        if self.keywords:
            for kw in self.keywords:
                if kw in session.msg_text:
                    break
            else:
                # no keyword matches
                return False

        return await self._check_perm(session)

    async def _check_perm(self, session: "NLPSession") -> bool:
        """
        Check if the session has sufficient permission to
        call the command.

        :param session: NLPSession object
        :return: the event has the permission
        """
        return await self.perm_checker_func(session.bot, session.event)


class NLPManager:
    """INTERNAL API"""

    _nl_processors: Set[NLProcessor] = set()

    def __init__(self):
        self.nl_processors = NLPManager._nl_processors.copy()

    @classmethod
    def add_nl_processor(cls, processor: NLProcessor) -> None:
        """Register a natural language processor

        Args:
            processor (NLProcessor): Processor object
        """
        if processor in cls._nl_processors:
            warnings.warn(f"NLProcessor {processor} already exists")
            return
        cls._nl_processors.add(processor)

    @classmethod
    def remove_nl_processor(cls, processor: NLProcessor) -> bool:
        """Remove a natural language processor globally

        Args:
            processor (NLProcessor): Processor to remove

        Returns:
            bool: Success or not
        """
        if processor in cls._nl_processors:
            cls._nl_processors.remove(processor)
            return True
        return False

    @classmethod
    def switch_nlprocessor_global(
        cls, processor: NLProcessor, state: Optional[bool] = None
    ) -> Optional[bool]:
        """Remove or add a natural language processor globally

        Args:
            processor (NLProcessor): Processor object

        Returns:
            bool: True if removed, False if added
        """
        if processor in cls._nl_processors and not state:
            cls._nl_processors.remove(processor)
            return True
        elif processor not in cls._nl_processors and state is not False:
            cls._nl_processors.add(processor)
            return False

    def switch_nlprocessor(
        self, processor: NLProcessor, state: Optional[bool] = None
    ) -> Optional[bool]:
        """Remove or add a natural language processor

        Args:
            processor (NLProcessor): Processor to remove

        Returns:
            bool: True if removed, False if added
        """
        if processor in self.nl_processors and not state:
            self.nl_processors.remove(processor)
            return True
        elif processor not in self.nl_processors and state is not False:
            self.nl_processors.add(processor)
            return False


class NLPSession(BaseSession):
    __slots__ = ("msg", "msg_text", "msg_images")

    def __init__(self, bot: NoneBot, event: CQEvent, msg: str):
        super().__init__(bot, event)
        self.msg = msg
        tmp_msg = Message(msg)
        self.msg_text = tmp_msg.extract_plain_text()
        self.msg_images = [
            s.data["url"] for s in tmp_msg if s.type == "image" and "url" in s.data
        ]


class NLPResult(NamedTuple):
    """
    Deprecated.
    Use class IntentCommand instead.
    """

    confidence: float
    cmd_name: Union[str, CommandName_T]
    cmd_args: Optional[CommandArgs_T] = None

    def to_intent_command(self):
        return IntentCommand(
            confidence=self.confidence, name=self.cmd_name, args=self.cmd_args
        )


class IntentCommand(NamedTuple):
    """
    To represent a command that we think the user may be intended to call.
    """

    confidence: float
    name: Union[str, CommandName_T]
    args: Optional[CommandArgs_T] = None
    current_arg: str = ""


async def handle_natural_language(
    bot: NoneBot, event: CQEvent, manager: NLPManager
) -> bool:
    """
    INTERNAL API

    Handle a message as natural language.

    This function is typically called by "handle_message".

    :param bot: NoneBot instance
    :param event: message event
    :param manager: natural language processor manager
    :return: the message is handled as natural language
    """
    session = NLPSession(bot, event, str(event.message))

    # use msg_text here because CQ code "share" may be very long,
    # at the same time some plugins may want to handle it
    msg_text_length = len(session.msg_text)

    # returns 1. processor result; 2. whether this processor is considered handled
    async def try_run_nlp(p: NLProcessor) -> Tuple[Any, bool]:
        try:
            should_run = await p.test(session, msg_text_length=msg_text_length)
            if should_run:
                return await p.func(session), True
            return None, False
        except Exception as e:
            logger.error(
                "An exception occurred while running "
                "some natural language processor:"
            )
            logger.exception(e)
            return None, True

    intent_commands: List[IntentCommand] = []
    procs_empty = True

    for res in asyncio.as_completed([try_run_nlp(p) for p in manager.nl_processors]):
        result, should_run = await res
        if not should_run:
            continue
        procs_empty = False
        if isinstance(result, NLPResult):
            intent_commands.append(result.to_intent_command())
        elif isinstance(result, IntentCommand):
            intent_commands.append(result)

    if procs_empty:
        return False

    intent_commands.sort(key=lambda ic: ic.confidence, reverse=True)
    logger.debug(f"Intent commands: {intent_commands}")

    if intent_commands and intent_commands[0].confidence >= 60.0:
        # choose the intent command with highest confidence
        chosen_cmd = intent_commands[0]
        logger.debug(f"Intent command with highest confidence: {chosen_cmd}")
        return (
            await call_command(
                bot,
                event,
                chosen_cmd.name,
                args=chosen_cmd.args,
                current_arg=chosen_cmd.current_arg,
                check_perm=False,
            )
            or False
        )
    logger.debug("No intent command has enough confidence")
    return False


__all__ = [
    "NLPSession",
    "NLPResult",
    "IntentCommand",
]



import asyncio
from functools import partial, singledispatchmethod
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Dict, Optional, TypeVar

from nonebot import get_app, get_asgi, get_bots
from nonebot.adapters.cqhttp import Bot as CQBot
from nonebot.adapters.cqhttp.event import Event as NoneBotEvent
from nonebot.adapters.cqhttp.message import Message
from nonebot.exception import ApiNotAvailable
from nonebot.plugin import on_message, on_metaevent, on_notice, on_request
from nonebot.typing import T_Handler

if TYPE_CHECKING:
    from nonetrip.comp.typing import Message_T

_AsyncCallable_T = TypeVar("_AsyncCallable_T", bound=Callable[..., Coroutine])
_HandlerDecorator = Callable[[_AsyncCallable_T], _AsyncCallable_T]


class Event(dict):
    """
    封装从 CQHTTP 收到的事件数据对象（字典），提供属性以获取其中的字段。
    除 `type` 和 `detail_type` 属性对于任何事件都有效外，其它属性存在与否（不存在则返回
    `None`）依事件不同而不同。
    """

    @staticmethod
    def from_payload(payload: NoneBotEvent) -> "Optional[Event]":
        """
        从 CQHTTP 事件数据构造 `Event` 对象。
        """
        try:
            e = Event(payload.dict())
            _ = e.type, e.detail_type
            return e
        except KeyError:
            return None

    @property
    def type(self) -> str:
        """
        事件类型，有 ``message``、``notice``、``request``、``meta_event`` 等。
        """
        return self["post_type"]

    @property
    def detail_type(self) -> str:
        """
        事件具体类型，依 `type` 的不同而不同，以 ``message`` 类型为例，有
        ``private``、``group``、``discuss`` 等。
        """
        return self[f"{self.type}_type"]

    @property
    def sub_type(self) -> Optional[str]:
        """
        事件子类型，依 `detail_type` 不同而不同，以 ``message.private`` 为例，有
        ``friend``、``group``、``discuss``、``other`` 等。
        """
        return self.get("sub_type")

    @property
    def name(self):
        """
        事件名，对于有 `sub_type` 的事件，为 ``{type}.{detail_type}.{sub_type}``，否则为
        ``{type}.{detail_type}``。
        """
        n = self.type + "." + self.detail_type
        if self.sub_type:
            n += "." + self.sub_type
        return n

    self_id: int  # 机器人自身 ID
    user_id: Optional[int]  # 用户 ID
    operator_id: Optional[int]  # 操作者 ID
    group_id: Optional[int]  # 群 ID
    discuss_id: Optional[int]  # 讨论组 ID
    message_id: Optional[int]  # 消息 ID
    message: Optional[Message]  # 消息
    raw_message: Optional[str]  # 未经 CQHTTP 处理的原始消息
    sender: Optional[Dict[str, Any]]  # 消息发送者信息
    anonymous: Optional[Dict[str, Any]]  # 匿名信息
    file: Optional[Dict[str, Any]]  # 文件信息
    comment: Optional[str]  # 请求验证消息
    flag: Optional[str]  # 请求标识

    def copy(self):
        return self.__class__(self.copy())

    def __getattr__(self, key) -> Optional[Any]:
        return self.get(key)

    def __setattr__(self, key, value) -> None:
        self[key] = value

    def __repr__(self) -> str:
        return f"<Event, {super().__repr__()}>"


class CQHttp:
    message_matcher = on_message()
    notice_handler = on_notice()
    request_handler = on_request()
    metaevent_handler = on_metaevent()

    @property
    def asgi(self):
        return get_asgi()

    @property
    def server_app(self):
        return get_app()

    @property
    def logger(self):
        from .comp.log import logger

        return logger

    @property
    def loop(self):
        return asyncio.get_running_loop()

    @property
    def bot(self) -> CQBot:
        for bot in get_bots().values():
            if not isinstance(bot, CQBot):
                continue
            return bot
        raise ApiNotAvailable("nonetrip")

    def _handler_factory(
        self,
        function: Callable[[Event], Coroutine],
        post_type: Optional[str] = None,
    ) -> T_Handler:
        async def handler(bot: CQBot, event: NoneBotEvent):
            if post_type is not None and event.post_type != post_type:
                return
            nonebot_event = Event.from_payload(event)
            if nonebot_event is None:
                return
            return await function(nonebot_event)

        return handler

    @singledispatchmethod
    def on_message(self, arg: _AsyncCallable_T) -> _AsyncCallable_T:
        self.message_matcher.append_handler(self._handler_factory(arg))
        return arg

    @on_message.register  # type:ignore
    def _on_specified_message(self, arg: str) -> _HandlerDecorator:
        def wrapper(function: _AsyncCallable_T) -> _AsyncCallable_T:
            self.message_matcher.append_handler(self._handler_factory(function, arg))
            return function

        return wrapper

    @singledispatchmethod
    def on_notice(self, arg):
        self.notice_handler.append_handler(self._handler_factory(arg))
        return arg

    @on_notice.register  # type: ignore
    def _on_specified_notice(self, arg: str) -> _HandlerDecorator:
        def wrapper(function: _AsyncCallable_T) -> _AsyncCallable_T:
            self.notice_handler.append_handler(self._handler_factory(function, arg))
            return function

        return wrapper

    @singledispatchmethod
    def on_request(self, arg: _AsyncCallable_T) -> _AsyncCallable_T:
        self.request_handler.append_handler(self._handler_factory(arg))
        return arg

    @on_request.register  # type: ignore
    def _on_specified_request(self, arg: str) -> _HandlerDecorator:
        def wrapper(function: _AsyncCallable_T) -> _AsyncCallable_T:
            self.request_handler.append_handler(self._handler_factory(function, arg))
            return function

        return wrapper

    @singledispatchmethod
    def on_metaevent(self, arg: _AsyncCallable_T) -> _AsyncCallable_T:
        self.metaevent_handler.append_handler(self._handler_factory(arg))
        return arg

    @on_metaevent.register  # type:ignore
    def _on_specified_metaevent(self, arg: str) -> _HandlerDecorator:
        def wrapper(function: _AsyncCallable_T) -> _AsyncCallable_T:
            self.metaevent_handler.append_handler(self._handler_factory(function, arg))
            return function

        return wrapper

    on_meta_event = on_metaevent

    async def send(self, event: Event, message: "Message_T", **kwargs):
        bot = get_bots().get(str(event.self_id))
        assert (bot is not None) and isinstance(bot, CQBot)
        return await bot.send(NoneBotEvent(**event), str(message), **kwargs)

    async def call_action(self, action: str, **kwargs):
        return await self.bot.call_api(action, **kwargs)

    def __getattr__(self, key: str) -> Callable[..., Coroutine]:
        return partial(self.call_action, key)



import asyncio
import re
from typing import Iterable, Set, Tuple

from nonebot.adapters.cqhttp.message import Message, MessageSegment
from nonebot.adapters.cqhttp.utils import escape, unescape

from ..poly import Event as CQEvent
from . import NoneBot
from .command import SwitchException, handle_command
from .log import logger
from .natural_language import handle_natural_language
from .plugin import PluginManager
from .typing import MessagePreprocessor_T

_message_preprocessors: Set[MessagePreprocessor_T] = set()


def message_preprocessor(func: MessagePreprocessor_T) -> MessagePreprocessor_T:
    _message_preprocessors.add(func)
    return func


class CanceledException(Exception):
    """
    Raised by message_preprocessor indicating that
    the bot should ignore the message
    """

    def __init__(self, reason):
        """
        :param reason: reason to ignore the message
        """
        self.reason = reason


async def handle_message(bot: NoneBot, event: CQEvent) -> None:
    """INTERNAL API"""
    _log_message(event)

    assert isinstance(event.message, Message)
    if not event.message:
        event.message.append(MessageSegment.text(""))  # type: ignore

    raw_to_me = event.get("to_me", False)
    _check_at_me(bot, event)
    _check_calling_me_nickname(bot, event)
    event["to_me"] = raw_to_me or event["to_me"]

    coros = []
    plugin_manager = PluginManager()
    for preprocessor in _message_preprocessors:
        coros.append(preprocessor(bot, event, plugin_manager))
    if coros:
        try:
            await asyncio.gather(*coros)
        except CanceledException as e:
            logger.info(f'Message {event["message_id"]} is ignored: {e.reason}')
            return

    while True:
        try:
            handled = await handle_command(bot, event, plugin_manager.cmd_manager)
            break
        except SwitchException as e:
            # we are sure that there is no session existing now
            event["message"] = e.new_message
            event["to_me"] = True
    if handled:
        logger.info(f"Message {event.message_id} is handled as a command")
        return

    handled = await handle_natural_language(bot, event, plugin_manager.nlp_manager)
    if handled:
        logger.info(f"Message {event.message_id} is handled " f"as natural language")
        return


def _check_at_me(bot: NoneBot, event: CQEvent) -> None:
    assert event.message is not None
    if event.detail_type == "private":
        event["to_me"] = True
    else:
        # group or discuss
        event["to_me"] = False
        at_me_seg = MessageSegment.at(event.self_id)

        # check the first segment
        first_msg_seg = event.message[0]
        if first_msg_seg == at_me_seg:
            event["to_me"] = True
            del event.message[0]

        if not event["to_me"]:
            # check the last segment
            i = -1
            last_msg_seg = event.message[i]
            if (
                last_msg_seg.type == "text"
                and not last_msg_seg.data["text"].strip()
                and len(event.message) >= 2
            ):
                i -= 1
                last_msg_seg = event.message[i]

            if last_msg_seg == at_me_seg:
                event["to_me"] = True
                del event.message[i:]

        if not event.message:
            event.message.append(MessageSegment.text(""))


def _check_calling_me_nickname(bot: NoneBot, event: CQEvent) -> None:
    first_msg_seg = event.message[0]  # type:ignore
    if first_msg_seg.type != "text":
        return

    first_text: str = first_msg_seg.data["text"]

    if bot.config.NICKNAME:
        nicknames: Tuple[str, ...]
        # check if the user is calling me with my nickname
        if isinstance(bot.config.NICKNAME, str) or not isinstance(
            bot.config.NICKNAME, Iterable
        ):
            nicknames = (bot.config.NICKNAME,)
        else:
            nicknames = tuple(filter(lambda n: n, bot.config.NICKNAME))
        nickname_regex = "|".join(nicknames)
        m = re.search(rf"^({nickname_regex})([\s,，]*|$)", first_text, re.IGNORECASE)
        if m:
            nickname = m.group(1)
            logger.debug(f"User is calling me {nickname}")
            event["to_me"] = True
            first_msg_seg.data["text"] = first_text[m.end() :]


def _log_message(event: CQEvent) -> None:
    msg_from = str(event.user_id)
    if event.detail_type == "group":
        msg_from += f"@[群:{event.group_id}]"
    elif event.detail_type == "discuss":
        msg_from += f"@[讨论组:{event.discuss_id}]"
    logger.info(
        f"Self: {event.self_id}, "
        f"Message {event.message_id} from {msg_from}: "
        f"{repr(str(event.message))}"
    )


__all__ = [
    "message_preprocessor",
    "CanceledException",
    "Message",
    "MessageSegment",
    "escape",
    "unescape",
]



from ..poly import Event as CQEvent
from . import NoneBot
from .helpers import send
from .typing import Message_T


class BaseSession:
    __slots__ = ("bot", "event")

    def __init__(self, bot: NoneBot, event: CQEvent):
        self.bot = bot
        self.event = event

    @property
    def ctx(self) -> CQEvent:
        return self.event

    @ctx.setter
    def ctx(self, val: CQEvent) -> None:
        self.event = val

    @property
    def self_id(self) -> int:
        return self.event.self_id

    async def send(
        self,
        message: Message_T,
        *,
        at_sender: bool = False,
        ensure_private: bool = False,
        ignore_failure: bool = True,
        **kwargs
    ) -> None:
        """
        Send a message ignoring failure by default.

        :param message: message to send
        :param at_sender: @ the sender if in group or discuss chat
        :param ensure_private: ensure the message is sent to private chat
        :param ignore_failure: if any CQHttpError raised, ignore it
        :return: the result returned by CQHTTP
        """
        return await send(
            self.bot,
            self.event,
            message,
            at_sender=at_sender,
            ensure_private=ensure_private,
            ignore_failure=ignore_failure,
            **kwargs
        )


__all__ = [
    "BaseSession",
]



"""
此模块提供事件总线相关类。
"""

import asyncio
from collections import defaultdict
from typing import Any, Callable, Coroutine, DefaultDict, Iterable, List, Set

AsyncCallable = Callable[..., Coroutine]


async def parallel_run(functions: Iterable[AsyncCallable], *args, **kwargs):
    return await asyncio.gather(*map(lambda f: f(*args, **kwargs), functions))


class EventBus:
    def __init__(self):
        self._subscribers: DefaultDict[str, Set[AsyncCallable]] = defaultdict(set)
        self._hooks_before: DefaultDict[str, Set[AsyncCallable]] = defaultdict(set)

    def subscribe(self, event: str, func: Callable) -> None:
        self._subscribers[event].add(func)

    def unsubscribe(self, event: str, func: Callable) -> None:
        if func in self._subscribers[event]:
            self._subscribers[event].remove(func)

    def hook_before(self, event: str, func: Callable) -> None:
        self._hooks_before[event].add(func)

    def unhook_before(self, event: str, func: Callable) -> None:
        if func in self._hooks_before[event]:
            self._hooks_before[event].remove(func)

    def on(self, event: str) -> Callable:
        def decorator(func: Callable) -> Callable:
            self.subscribe(event, func)
            return func

        return decorator

    def before(self, event: str) -> Callable:
        def decorator(func: Callable) -> Callable:
            self.hook_before(event, func)
            return func

        return decorator

    async def emit(self, event: str, *args, **kwargs) -> List[Any]:
        event_copy = event

        while True:
            await parallel_run(self._hooks_before[event], *args, **kwargs)
            event, *sub_event = event.rsplit(".", maxsplit=1)
            if not sub_event:
                # the current event is the root event
                break
        event = event_copy

        results = []
        while True:
            results += await parallel_run(self._subscribers[event], *args, **kwargs)
            event, *sub_event = event.rsplit(".", maxsplit=1)
            if not sub_event:
                # the current event is the root event
                break
        return results



import re
from typing import List

from nonebot.adapters.cqhttp.message import Message
from nonetrip.comp.typing import Message_T


def _extract_text(arg: Message_T) -> str:
    """Extract all plain text segments from a message-like object."""
    arg_as_msg = Message(arg)
    return arg_as_msg.extract_plain_text()


def _extract_image_urls(arg: Message_T) -> List[str]:
    """Extract all image urls from a message-like object."""
    arg_as_msg = Message(arg)
    return [s.data["url"] for s in arg_as_msg if s.type == "image" and "url" in s.data]


def _extract_numbers(arg: Message_T) -> List[float]:
    """Extract all numbers (integers and floats) from a message-like object."""
    s = str(arg)
    return list(map(float, re.findall(r"[+-]?(\d*\.?\d+|\d+\.?\d*)", s)))


extract_text = _extract_text
extract_image_urls = _extract_image_urls
extract_numbers = _extract_numbers


__all__ = [
    "extract_text",
    "extract_image_urls",
    "extract_numbers",
]



"""
This module replicates the standard  modules already defined
in nonetrip.comp. This, yet another implementation, is experimental and may
be easier or harder to use than the standard one.
"""

from datetime import timedelta
from functools import partial
from typing import Callable, Iterable, Optional, Type, Union, overload

from nonetrip.comp.command import CommandSession
from nonetrip.comp.experimental.permission import (
    RoleCheckPolicy,
    aggregate_policy,
    check_permission,
)
from nonetrip.comp.plugin import on_command_custom, on_natural_language_custom
from nonetrip.comp.typing import (
    CommandHandler_T,
    CommandName_T,
    Patterns_T,
    NLPHandler_T,
)


def on_command(
    name: Union[str, CommandName_T],
    *,
    aliases: Union[Iterable[str], str] = (),
    patterns: Patterns_T = (),
    permission: Union[RoleCheckPolicy, Iterable[RoleCheckPolicy]] = lambda _: True,
    only_to_me: bool = True,
    privileged: bool = False,
    shell_like: bool = False,
    expire_timeout: Optional[timedelta] = ...,
    run_timeout: Optional[timedelta] = ...,
    session_class: Optional[Type[CommandSession]] = None
) -> Callable[[CommandHandler_T], CommandHandler_T]:
    """
    Decorator to register a function as a command.

    This function's description is consistent with nonetrip.comp.plugin.on_command.
    """
    if isinstance(permission, Iterable):
        permission = aggregate_policy(permission)
    perm_checker = partial(check_permission, policy=permission)
    return on_command_custom(
        name,
        aliases=aliases,
        patterns=patterns,
        only_to_me=only_to_me,
        privileged=privileged,
        shell_like=shell_like,
        perm_checker=perm_checker,
        expire_timeout=expire_timeout,
        run_timeout=run_timeout,
        session_class=session_class,
    )


@overload
def on_natural_language(func: NLPHandler_T) -> NLPHandler_T:
    """
    Decorator to register a function as a natural language processor with
    default kwargs.
    """


@overload
def on_natural_language(
    keywords: Optional[Union[Iterable[str], str]] = ...,
    *,
    permission: Union[RoleCheckPolicy, Iterable[RoleCheckPolicy]] = ...,
    only_to_me: bool = ...,
    only_short_message: bool = ...,
    allow_empty_message: bool = ...
) -> Callable[[NLPHandler_T], NLPHandler_T]:
    """
    Decorator to register a function as a natural language processor.

    This function's description is consistent with nonetrip.comp.plugin.on_natural_language,
    except permission, where you pass a function instead of an int.
    """


def on_natural_language(
    keywords: Union[Optional[Iterable[str]], str, NLPHandler_T] = None,
    *,
    permission: Union[RoleCheckPolicy, Iterable[RoleCheckPolicy]] = lambda _: True,
    only_to_me: bool = True,
    only_short_message: bool = True,
    allow_empty_message: bool = False
):
    """
    Implementation of on_natural_language overloads.
    """
    if isinstance(permission, Iterable):
        permission = aggregate_policy(permission)
    perm_checker = partial(check_permission, policy=permission)
    return on_natural_language_custom(
        keywords,
        only_to_me=only_to_me,
        only_short_message=only_short_message,
        allow_empty_message=allow_empty_message,
        perm_checker=perm_checker,
    )


# command groups not implemented yet

__all__ = ["on_command", "on_natural_language"]



import re
from typing import Callable, Any

from nonetrip.comp.typing import Filter_T
from nonetrip.comp.command.argfilter import ValidateError


class BaseValidator:
    """INTERNAL API"""

    def __init__(self, message=None):
        self.message = message

    def raise_failure(self):
        raise ValidateError(self.message)


def _raise_failure(message):
    raise ValidateError(message)


def not_empty(message=None) -> Filter_T:
    """
    Validate any object to ensure it's not empty (is None or has no elements).
    """

    def validate(value):
        if value is None:
            _raise_failure(message)
        if hasattr(value, "__len__") and value.__len__() == 0:
            _raise_failure(message)
        return value

    return validate


def fit_size(min_length: int = 0, max_length: int = None, message=None) -> Filter_T:
    """
    Validate any sized object to ensure the size/length
    is in a given range [min_length, max_length].
    """

    def validate(value):
        length = len(value) if value is not None else 0
        if length < min_length or (max_length is not None and length > max_length):
            _raise_failure(message)
        return value

    return validate


def match_regex(
    pattern: str, message=None, *, flags=0, fullmatch: bool = False
) -> Filter_T:
    """
    Validate any string object to ensure it matches a given pattern.
    """

    pattern_ = re.compile(pattern, flags)

    def validate(value):
        if fullmatch:
            if not re.fullmatch(pattern_, value):
                _raise_failure(message)
        else:
            if not re.match(pattern_, value):
                _raise_failure(message)
        return value

    return validate


def ensure_true(bool_func: Callable[[Any], bool], message=None) -> Filter_T:
    """
    Validate any object to ensure the result of applying
    a boolean function to it is True.
    """

    def validate(value):
        if bool_func(value) is not True:
            _raise_failure(message)
        return value

    return validate


def between_inclusive(start=None, end=None, message=None) -> Filter_T:
    """
    Validate any comparable object to ensure it's between
    `start` and `end` inclusively.
    """

    def validate(value):
        if start is not None and value < start:
            _raise_failure(message)
        if end is not None and end < value:
            _raise_failure(message)
        return value

    return validate


__all__ = [
    "not_empty",
    "fit_size",
    "match_regex",
    "ensure_true",
    "between_inclusive",
]



from typing import Any, Dict, NamedTuple, Optional

from aiocache import cached

from ..poly import Event as CQEvent
from . import NoneBot
from .exceptions import CQHttpError

PRIVATE_FRIEND = 0x0001
PRIVATE_GROUP = 0x0002
PRIVATE_DISCUSS = 0x0004
PRIVATE_OTHER = 0x0008
PRIVATE = 0x000F
DISCUSS = 0x00F0
GROUP_MEMBER = 0x0100
GROUP_ADMIN = 0x0200
GROUP_OWNER = 0x0400
GROUP = 0x0F00
SUPERUSER = 0xF000
EVERYBODY = 0xFFFF

IS_NOBODY = 0x0000
IS_PRIVATE_FRIEND = PRIVATE_FRIEND
IS_PRIVATE_GROUP = PRIVATE_GROUP
IS_PRIVATE_DISCUSS = PRIVATE_DISCUSS
IS_PRIVATE_OTHER = PRIVATE_OTHER
IS_PRIVATE = PRIVATE
IS_DISCUSS = DISCUSS
IS_GROUP_MEMBER = GROUP_MEMBER
IS_GROUP_ADMIN = GROUP_MEMBER | GROUP_ADMIN
IS_GROUP_OWNER = GROUP_ADMIN | GROUP_OWNER
IS_GROUP = GROUP
IS_SUPERUSER = 0xFFFF

_min_event_fields = (
    "self_id",
    "message_type",
    "sub_type",
    "user_id",
    "discuss_id",
    "group_id",
    "anonymous",
)


class _MinEvent(NamedTuple):
    self_id: int
    message_type: str
    sub_type: str
    user_id: int
    discuss_id: int
    group_id: int
    anonymous: Optional[Dict[str, Any]] = None


async def check_permission(
    bot: NoneBot, event: CQEvent, permission_required: int
) -> bool:
    """
    Check if the event context has the permission required.

    :param bot: NoneBot instance
    :param event: message event
    :param permission_required: permission required
    :return: the context has the permission
    """
    min_event_kwargs = {}
    for field in _min_event_fields:
        if field in event:
            min_event_kwargs[field] = event[field]
        else:
            min_event_kwargs[field] = None
    min_event = _MinEvent(**min_event_kwargs)
    return await _check(bot, min_event, permission_required)


@cached(ttl=2 * 60)  # cache the result for 2 minute
async def _check(bot: NoneBot, min_event: _MinEvent, permission_required: int) -> bool:
    permission = 0
    if min_event.user_id in bot.config.SUPERUSERS:
        permission |= IS_SUPERUSER
    if min_event.message_type == "private":
        if min_event.sub_type == "friend":
            permission |= IS_PRIVATE_FRIEND
        elif min_event.sub_type == "group":
            permission |= IS_PRIVATE_GROUP
        elif min_event.sub_type == "discuss":
            permission |= IS_PRIVATE_DISCUSS
        elif min_event.sub_type == "other":
            permission |= IS_PRIVATE_OTHER
    elif min_event.message_type == "group":
        permission |= IS_GROUP_MEMBER
        if not min_event.anonymous:
            try:
                member_info = await bot.get_group_member_info(
                    self_id=min_event.self_id,
                    group_id=min_event.group_id,
                    user_id=min_event.user_id,
                    no_cache=True,
                )
                if member_info:
                    if member_info["role"] == "owner":
                        permission |= IS_GROUP_OWNER
                    elif member_info["role"] == "admin":
                        permission |= IS_GROUP_ADMIN
            except CQHttpError:
                pass
    elif min_event.message_type == "discuss":
        permission |= IS_DISCUSS

    return bool(permission & permission_required)


__all__ = [
    "PRIVATE_FRIEND",
    "PRIVATE_GROUP",
    "PRIVATE_DISCUSS",
    "PRIVATE_OTHER",
    "PRIVATE",
    "DISCUSS",
    "GROUP_MEMBER",
    "GROUP_ADMIN",
    "GROUP_OWNER",
    "GROUP",
    "SUPERUSER",
    "EVERYBODY",
    "check_permission",
]



import importlib
import os
import re
import shlex
import sys
import warnings
from datetime import timedelta
from functools import partial
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    overload,
)

from nonetrip.comp import permission as perm

from .command import Command, CommandManager, CommandSession
from .log import logger
from .natural_language import NLPManager, NLProcessor
from .notice_request import EventHandler, _bus
from .typing import (
    CommandHandler_T,
    CommandName_T,
    NLPHandler_T,
    NoticeHandler_T,
    Patterns_T,
    PermChecker_T,
    RequestHandler_T,
)


class Plugin:
    __slots__ = (
        "module",
        "name",
        "usage",
        "commands",
        "nl_processors",
        "event_handlers",
    )

    def __init__(
        self,
        module: ModuleType,
        name: Optional[str] = None,
        usage: Optional[Any] = None,
        commands: Optional[Set[Command]] = None,
        nl_processors: Optional[Set[NLProcessor]] = None,
        event_handlers: Optional[Set[EventHandler]] = None,
    ):
        """Creates a plugin with no name, no usage, and no handlers."""

        self.module = module
        self.name = name
        self.usage = usage
        self.commands: Set[Command] = commands or set()
        self.nl_processors: Set[NLProcessor] = nl_processors or set()
        self.event_handlers: Set[EventHandler] = event_handlers or set()

    class GlobalTemp:
        """INTERNAL API"""

        commands: Set[Command] = set()
        nl_processors: Set[NLProcessor] = set()
        event_handlers: Set[EventHandler] = set()

        @classmethod
        def clear(cls):
            cls.commands.clear()
            cls.nl_processors.clear()
            cls.event_handlers.clear()

        @classmethod
        def make_plugin(cls, module: ModuleType):
            return Plugin(
                module=module,
                name=getattr(module, "__plugin_name__", None),
                usage=getattr(module, "__plugin_usage__", None),
                commands={*cls.commands},
                nl_processors={*cls.nl_processors},
                event_handlers={*cls.event_handlers},
            )


class PluginManager:
    _plugins: Dict[str, Plugin] = {}

    def __init__(self):
        self.cmd_manager = CommandManager()
        self.nlp_manager = NLPManager()

    @classmethod
    def add_plugin(cls, module_path: str, plugin: Plugin) -> None:
        """Register a plugin

        Args:
            module_path (str): module path
            plugin (Plugin): Plugin object
        """
        if module_path in cls._plugins:
            warnings.warn(f"Plugin {module_path} already exists")
            return
        cls._plugins[module_path] = plugin

    @classmethod
    def get_plugin(cls, module_path: str) -> Optional[Plugin]:
        """Get plugin object by plugin module path

        Args:
            module_path (str): Plugin module path

        Returns:
            Optional[Plugin]: Plugin object
        """
        return cls._plugins.get(module_path, None)

    @classmethod
    def remove_plugin(cls, module_path: str) -> bool:
        """Remove a plugin by plugin module path

        ** Warning: This function not remove plugin actually! **
        ** Just remove command, nlprocessor and event handlers **

        Args:
            module_path (str): Plugin module path

        Returns:
            bool: Success or not
        """
        plugin = cls.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not exists")
            return False
        for command in plugin.commands:
            CommandManager.remove_command(command.name)
        for nl_processor in plugin.nl_processors:
            NLPManager.remove_nl_processor(nl_processor)
        for event_handler in plugin.event_handlers:
            for event in event_handler.events:
                _bus.unsubscribe(event, event_handler.func)
        del cls._plugins[module_path]
        return True

    @classmethod
    def switch_plugin_global(
        cls, module_path: str, state: Optional[bool] = None
    ) -> None:
        """Change plugin state globally or simply switch it if `state` is None

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = cls.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for command in plugin.commands:
            CommandManager.switch_command_global(command.name, state)
        for nl_processor in plugin.nl_processors:
            NLPManager.switch_nlprocessor_global(nl_processor, state)
        for event_handler in plugin.event_handlers:
            for event in event_handler.events:
                if event_handler.func in _bus._subscribers[event] and not state:
                    _bus.unsubscribe(event, event_handler.func)
                elif (
                    event_handler.func not in _bus._subscribers[event]
                    and state is not False
                ):
                    _bus.subscribe(event, event_handler.func)

    @classmethod
    def switch_command_global(
        cls, module_path: str, state: Optional[bool] = None
    ) -> None:
        """Change plugin command state globally or simply switch it if `state` is None

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = cls.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for command in plugin.commands:
            CommandManager.switch_command_global(command.name, state)

    @classmethod
    def switch_nlprocessor_global(
        cls, module_path: str, state: Optional[bool] = None
    ) -> None:
        """Change plugin nlprocessor state globally or simply switch it if `state` is None

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = cls.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for processor in plugin.nl_processors:
            NLPManager.switch_nlprocessor_global(processor, state)

    @classmethod
    def switch_eventhandler_global(
        cls, module_path: str, state: Optional[bool] = None
    ) -> None:
        """Change plugin event handler state globally or simply switch it if `state` is None

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = cls.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for event_handler in plugin.event_handlers:
            for event in event_handler.events:
                if event_handler.func in _bus._subscribers[event] and not state:
                    _bus.unsubscribe(event, event_handler.func)
                elif (
                    event_handler.func not in _bus._subscribers[event]
                    and state is not False
                ):
                    _bus.subscribe(event, event_handler.func)

    def switch_plugin(self, module_path: str, state: Optional[bool] = None) -> None:
        """Change plugin state or simply switch it if `state` is None

        Tips:
            This method will only change the state of the plugin's
            commands and natural language processors since change
            state of the event handler for message is meaningless.

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = self.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for command in plugin.commands:
            self.cmd_manager.switch_command(command.name, state)
        for nl_processor in plugin.nl_processors:
            self.nlp_manager.switch_nlprocessor(nl_processor, state)

    def switch_command(self, module_path: str, state: Optional[bool] = None) -> None:
        """Change plugin command state or simply switch it if `state` is None

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = self.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for command in plugin.commands:
            self.cmd_manager.switch_command(command.name, state)

    def switch_nlprocessor(
        self, module_path: str, state: Optional[bool] = None
    ) -> None:
        """Change plugin nlprocessor state or simply switch it if `state` is None

        Args:
            module_path (str): Plugin module path
            state (Optional[bool]): State to change to. Defaults to None.
        """
        plugin = self.get_plugin(module_path)
        if not plugin:
            warnings.warn(f"Plugin {module_path} not found")
            return
        for processor in plugin.nl_processors:
            self.nlp_manager.switch_nlprocessor(processor, state)


def load_plugin(module_path: str) -> Optional[Plugin]:
    """Load a module as a plugin

    Args:
        module_path (str): path of module to import

    Returns:
        Optional[Plugin]: Plugin object loaded
    """
    Plugin.GlobalTemp.clear()
    try:
        module = importlib.import_module(module_path)
        plugin = Plugin.GlobalTemp.make_plugin(module)
        PluginManager.add_plugin(module_path, plugin)
        logger.info(f'Succeeded to import "{module_path}"')
        return plugin
    except Exception as e:
        logger.error(f'Failed to import "{module_path}", error: {e}')
        logger.exception(e)
        return None


def reload_plugin(module_path: str) -> Optional[Plugin]:
    result = PluginManager.remove_plugin(module_path)
    if not result:
        return None

    for existed_module in list(
        filter(lambda x: x.startswith(module_path), sys.modules.keys())
    ):
        del sys.modules[existed_module]

    Plugin.GlobalTemp.clear()
    module: ModuleType
    try:
        module = importlib.import_module(module_path)
        plugin = Plugin.GlobalTemp.make_plugin(module)
        PluginManager.add_plugin(module_path, plugin)
        logger.info(f'Succeeded to reload "{module_path}"')
        return plugin
    except Exception as e:
        logger.error(f'Failed to reload "{module_path}", error: {e}')
        logger.exception(e)
        return None


def load_plugins(plugin_dir: str, module_prefix: str) -> Set[Plugin]:
    """Find all non-hidden modules or packages in a given directory,
    and import them with the given module prefix.

    Args:
        plugin_dir (str): Plugin directory to search
        module_prefix (str): Module prefix used while importing

    Returns:
        Set[Plugin]: Set of plugin objects successfully loaded
    """

    count = set()
    for name in os.listdir(plugin_dir):
        path = os.path.join(plugin_dir, name)
        if os.path.isfile(path) and (name.startswith("_") or not name.endswith(".py")):
            continue
        if os.path.isdir(path) and (
            name.startswith("_")
            or not os.path.exists(os.path.join(path, "__init__.py"))
        ):
            continue

        m = re.match(r"([_A-Z0-9a-z]+)(.py)?", name)
        if not m:
            continue

        result = load_plugin(f"{module_prefix}.{m.group(1)}")
        if result:
            count.add(result)
    return count


def load_builtin_plugins() -> Set[Plugin]:
    """
    Load built-in plugins distributed along with "nonebot" package.
    """
    plugin_dir = os.path.join(os.path.dirname(__file__), "plugins")
    return load_plugins(plugin_dir, "nonetrip.comp.plugins")


def get_loaded_plugins() -> Set[Plugin]:
    """
    Get all plugins loaded.

    :return: a set of Plugin objects
    """
    return set(PluginManager._plugins.values())


def on_command_custom(
    name: Union[str, CommandName_T],
    *,
    aliases: Union[Iterable[str], str],
    patterns: Patterns_T,
    only_to_me: bool,
    privileged: bool,
    shell_like: bool,
    perm_checker: PermChecker_T,
    expire_timeout: Optional[timedelta],
    run_timeout: Optional[timedelta],
    session_class: Optional[Type[CommandSession]],
) -> Callable[[CommandHandler_T], CommandHandler_T]:
    """
    INTERNAL API

    The implementation of on_command function with custom per checker function.
    dev: This function may not last long. Kill it when this function is referenced
    only once
    """

    def deco(func: CommandHandler_T) -> CommandHandler_T:
        if not isinstance(name, (str, tuple)):
            raise TypeError("the name of a command must be a str or tuple")
        if not name:
            raise ValueError("the name of a command must not be empty")
        if session_class is not None and not issubclass(session_class, CommandSession):
            raise TypeError("session_class must be a subclass of CommandSession")

        cmd_name = (name,) if isinstance(name, str) else name

        cmd = Command(
            name=cmd_name,
            func=func,
            only_to_me=only_to_me,
            privileged=privileged,
            perm_checker_func=perm_checker,
            expire_timeout=expire_timeout,
            run_timeout=run_timeout,
            session_class=session_class,
        )

        if shell_like:

            async def shell_like_args_parser(session):
                session.args["argv"] = shlex.split(session.current_arg)

            cmd.args_parser_func = shell_like_args_parser

        CommandManager.add_command(cmd_name, cmd)
        CommandManager.add_aliases(aliases, cmd)
        CommandManager.add_patterns(patterns, cmd)

        Plugin.GlobalTemp.commands.add(cmd)
        setattr(func, "args_parser", cmd.args_parser)

        return func

    return deco


def on_command(
    name: Union[str, CommandName_T],
    *,
    aliases: Union[Iterable[str], str] = (),
    patterns: Patterns_T = (),
    permission: int = perm.EVERYBODY,
    only_to_me: bool = True,
    privileged: bool = False,
    shell_like: bool = False,
    expire_timeout: Optional[timedelta] = None,
    run_timeout: Optional[timedelta] = None,
    session_class: Optional[Type[CommandSession]] = None,
) -> Callable[[CommandHandler_T], CommandHandler_T]:
    """
    Decorator to register a function as a command.

    :param name: command name (e.g. 'echo' or ('random', 'number'))
    :param aliases: aliases of command name, for convenient access
    :param patterns: custom regex pattern for the command.
           Please use this carefully. Abuse may cause performance problem.
           Also, Please notice that if a message is matched by this method,
           it will use the full command as session current_arg.
    :param permission: permission required by the command
    :param only_to_me: only handle messages to me
    :param privileged: can be run even when there is already a session
    :param shell_like: use shell-like syntax to split arguments
    :param expire_timeout: will override SESSION_EXPIRE_TIMEOUT if provided
    :param run_timeout: will override SESSION_RUN_TIMEOUT if provided
    :param session_class: session class
    """
    perm_checker = partial(perm.check_permission, permission_required=permission)
    return on_command_custom(
        name,
        aliases=aliases,
        patterns=patterns,
        only_to_me=only_to_me,
        privileged=privileged,
        shell_like=shell_like,
        perm_checker=perm_checker,
        expire_timeout=expire_timeout,
        run_timeout=run_timeout,
        session_class=session_class,
    )


def on_natural_language_custom(
    keywords: Union[Optional[Iterable[str]], str, NLPHandler_T],
    *,
    only_to_me: bool,
    only_short_message: bool,
    allow_empty_message: bool,
    perm_checker: PermChecker_T,
) -> Union[Callable[[NLPHandler_T], NLPHandler_T], NLPHandler_T]:
    """
    INTERNAL API

    The implementation of on_natural_language function with custom per checker function.
    dev: This function may not last long. Kill it when this function is referenced
    only once
    """

    def deco(func: NLPHandler_T) -> NLPHandler_T:
        nl_processor = NLProcessor(
            func=func,
            keywords=keywords,  # type: ignore
            only_to_me=only_to_me,
            only_short_message=only_short_message,
            allow_empty_message=allow_empty_message,
            perm_checker_func=perm_checker,
        )

        NLPManager.add_nl_processor(nl_processor)
        Plugin.GlobalTemp.nl_processors.add(nl_processor)
        return func

    if callable(keywords):
        # here "keywords" is the function to be decorated
        # applies default args provided by this function
        return on_natural_language()(keywords)
    else:
        if isinstance(keywords, str):
            keywords = (keywords,)
        return deco


@overload
def on_natural_language(func: NLPHandler_T) -> NLPHandler_T:
    """
    Decorator to register a function as a natural language processor with
    default kwargs.
    """


@overload
def on_natural_language(
    keywords: Optional[Union[Iterable[str], str]] = ...,
    *,
    permission: int = ...,
    only_to_me: bool = ...,
    only_short_message: bool = ...,
    allow_empty_message: bool = ...,
) -> Callable[[NLPHandler_T], NLPHandler_T]:
    """
    Decorator to register a function as a natural language processor.

    :param keywords: keywords to respond to, if None, respond to all messages
    :param permission: permission required by the processor
    :param only_to_me: only handle messages to me
    :param only_short_message: only handle short messages
    :param allow_empty_message: handle empty messages
    """


def on_natural_language(  # type: ignore
    keywords: Union[Optional[Iterable[str]], str, NLPHandler_T] = None,
    *,
    permission: int = perm.EVERYBODY,
    only_to_me: bool = True,
    only_short_message: bool = True,
    allow_empty_message: bool = False,
):
    """
    Implementation of on_natural_language overloads.
    """
    perm_checker = partial(perm.check_permission, permission_required=permission)
    return on_natural_language_custom(
        keywords,
        only_to_me=only_to_me,
        only_short_message=only_short_message,
        allow_empty_message=allow_empty_message,
        perm_checker=perm_checker,
    )


_Teh = TypeVar("_Teh", NoticeHandler_T, RequestHandler_T)


def _make_event_deco(post_type: str):
    def deco_deco(
        arg: Optional[Union[str, _Teh]] = None, *events: str
    ) -> Union[Callable[[_Teh], _Teh], _Teh]:
        def deco(func: _Teh) -> _Teh:
            if isinstance(arg, str):
                events_tmp = list(
                    map(lambda x: f"{post_type}.{x}", [arg, *events])
                )  # if arg is part of events str
                for e in events_tmp:
                    _bus.subscribe(e, func)  # type: ignore
                handler = EventHandler(events_tmp, func)
            else:
                _bus.subscribe(post_type, func)  # type: ignore
                handler = EventHandler([post_type], func)
            Plugin.GlobalTemp.event_handlers.add(handler)
            return func

        if callable(arg):
            return deco(arg)
        return deco

    return deco_deco


@overload  # type: ignore
def on_notice(func: NoticeHandler_T) -> NoticeHandler_T:
    ...


@overload
def on_notice(*events: str) -> Callable[[NoticeHandler_T], NoticeHandler_T]:
    ...


on_notice = _make_event_deco("notice")  # type: ignore[override]


@overload  # type: ignore
def on_request(func: RequestHandler_T) -> RequestHandler_T:
    ...


@overload
def on_request(*events: str) -> Callable[[RequestHandler_T], RequestHandler_T]:
    ...


on_request = _make_event_deco("request")  # type: ignore[override]


__all__ = [
    "Plugin",
    "PluginManager",
    "load_plugin",
    "reload_plugin",
    "load_plugins",
    "load_builtin_plugins",
    "get_loaded_plugins",
    "on_command",
    "on_natural_language",
    "on_notice",
    "on_request",
]



import asyncio
import re
import warnings
from datetime import datetime, timedelta
from functools import partial
from typing import (
    Any,
    Awaitable,
    Dict,
    Iterable,
    List,
    NoReturn,
    Optional,
    Pattern,
    Tuple,
    Type,
    Union,
)

from nonebot.adapters.cqhttp.message import Message
from nonetrip.comp import NoneBot
from nonetrip.comp.command.argfilter import ValidateError
from nonetrip.comp.helpers import context_id, render_expression, send
from nonetrip.comp.log import logger
from nonetrip.comp.session import BaseSession
from nonetrip.comp.typing import (
    CommandArgs_T,
    CommandHandler_T,
    CommandName_T,
    Filter_T,
    Message_T,
    Patterns_T,
    PermChecker_T,
    State_T,
)

from ...poly import Event as CQEvent

# key: context id
# value: CommandSession object
_sessions = {}  # type: Dict[str, "CommandSession"]


class CommandInterrupt(Exception):
    """INTERNAL API"""

    pass


class _YieldException(CommandInterrupt):
    """
    Raised by command.run(), session indicating that the waiting session
    will resume and current execution path should return immediately.
    """

    pass


class _PauseException(CommandInterrupt):
    """
    Raised by session.pause() indicating that the command session
    should be paused to ask the user for some arguments.
    """

    pass


class _FinishException(CommandInterrupt):
    """
    Raised by session.finish() indicating that the command session
    should be stopped and removed.
    """

    def __init__(self, result: bool = True):
        """
        :param result: succeeded to call the command
        """
        self.result = result


class SwitchException(CommandInterrupt):
    """
    INTERNAL API

    Raised by session.switch() indicating that the command session
    should be stopped and replaced with a new one (going through
    handle_message() again).

    Since the new message will go through handle_message() again,
    the later function should be notified. So this exception is
    intended to be propagated to handle_message().
    """

    def __init__(self, new_message: Message):
        """
        :param new_message: new message which should be placed in event
        """
        self.new_message = new_message


class Command:
    """
    INTERNAL API

    Note ... (Ellipsis) is a valid value for field expire_timeout and run_timeout.
    However I cannot type it until Python 3.10. see bugs.python.org/issue41810.
    """

    __slots__ = (
        "name",
        "func",
        "only_to_me",
        "privileged",
        "args_parser_func",
        "perm_checker_func",
        "expire_timeout",
        "run_timeout",
        "session_class",
    )

    def __init__(
        self,
        *,
        name: CommandName_T,
        func: CommandHandler_T,
        only_to_me: bool,
        privileged: bool,
        perm_checker_func: PermChecker_T,
        expire_timeout: Optional[timedelta],
        run_timeout: Optional[timedelta],
        session_class: Optional[Type["CommandSession"]],
    ):
        self.name = name
        self.func = func
        self.only_to_me = only_to_me
        self.privileged = privileged
        self.args_parser_func: Optional[CommandHandler_T] = None
        self.perm_checker_func = perm_checker_func  # returns True if can trigger
        self.expire_timeout = expire_timeout  # includes EllipsisType
        self.run_timeout = run_timeout  # includes EllipsisType
        self.session_class = session_class

    async def run(
        self, session: "CommandSession", *, check_perm: bool = True, dry: bool = False
    ) -> bool:
        """
        Run the command in a given session.

        :param session: CommandSession object
        :param check_perm: should check permission before running
        :param dry: just check any prerequisite, without actually running
        :return: the command is finished (or can be run, given dry == True)
        """
        has_perm = await self._check_perm(session) if check_perm else True
        if self.func and has_perm:
            if dry:
                return True

            if (
                session.current_arg_filters is not None
                and session.current_key is not None
            ):
                # argument-level filters are given, use them
                arg = session.current_arg
                config = session.bot.config
                for f in session.current_arg_filters:
                    try:
                        res = f(arg)
                        if isinstance(res, Awaitable):
                            res = await res
                        arg = res
                    except ValidateError as e:
                        # validation failed
                        if config.MAX_VALIDATION_FAILURES > 0:
                            # should check number of validation failures
                            session.state["__validation_failure_num"] = (
                                session.state.get("__validation_failure_num", 0) + 1
                            )

                            if (
                                session.state["__validation_failure_num"]
                                >= config.MAX_VALIDATION_FAILURES
                            ):
                                # noinspection PyProtectedMember
                                session.finish(
                                    render_expression(
                                        config.TOO_MANY_VALIDATION_FAILURES_EXPRESSION
                                    ),
                                    **session._current_send_kwargs,
                                )

                        failure_message = e.message
                        if failure_message is None:
                            failure_message = render_expression(
                                config.DEFAULT_VALIDATION_FAILURE_EXPRESSION
                            )
                        # noinspection PyProtectedMember
                        session.pause(failure_message, **session._current_send_kwargs)

                # passed all filters
                session.state[session.current_key] = arg
            else:
                # fallback to command-level args_parser_func
                if self.args_parser_func:
                    await self.args_parser_func(session)
                if (
                    session.current_key is not None
                    and session.current_key not in session.state
                ):
                    # args_parser_func didn't set state, here we set it
                    session.state[session.current_key] = session.current_arg
            if session.waiting:
                session._future.set_result(True)
                raise _YieldException()
            else:
                await self.func(session)
            return True
        return False

    async def _check_perm(self, session: "CommandSession") -> bool:
        """
        Check if the session has sufficient permission to
        call the command.

        :param session: CommandSession object
        :return: the session has the permission
        """
        return await self.perm_checker_func(session.bot, session.event)

    def args_parser(self, parser_func: CommandHandler_T) -> CommandHandler_T:
        """
        Decorator to register a function as the arguments parser of
        the corresponding command.
        """
        self.args_parser_func = parser_func
        return parser_func

    def __repr__(self):
        return f"<Command, name={self.name.__repr__()}>"

    def __str__(self):
        return self.__repr__()


class CommandManager:
    """Global Command Manager"""

    _commands = {}  # type: Dict[CommandName_T, Command]
    _aliases = {}  # type: Dict[str, Command]
    _switches = {}  # type: Dict[Command, bool]
    _patterns = {}  # type: Dict[Pattern[str], Command]

    def __init__(self):
        self.commands = CommandManager._commands.copy()
        self.aliases = CommandManager._aliases.copy()
        self.switches = CommandManager._switches.copy()
        self.patterns = CommandManager._patterns.copy()

    @classmethod
    def add_command(cls, cmd_name: CommandName_T, cmd: Command) -> None:
        """Register a command

        Args:
            cmd_name (CommandName_T): Command name
            cmd (Command): Command object
        """
        if cmd_name in cls._commands:
            warnings.warn(f"Command {cmd_name} already exists")
            return
        cls._switches[cmd] = True
        cls._commands[cmd_name] = cmd

    @classmethod
    def reload_command(cls, cmd_name: CommandName_T, cmd: Command) -> None:
        """Reload a command

        **Warning! Dangerous function**

        Args:
            cmd_name (CommandName_T): Command name
            cmd (Command): Command object
        """
        if cmd_name not in cls._commands:
            warnings.warn(
                f"Command {cmd_name} does not exist. Please use add_command instead"
            )
            return

        cmd_ = cls._commands[cmd_name]
        if cmd_ in cls._switches:
            del cls._switches[cmd_]

        cls._switches[cmd] = True
        cls._commands[cmd_name] = cmd

    @classmethod
    def remove_command(cls, cmd_name: CommandName_T) -> bool:
        """Remove a command

        **Warning! Dangerous function**

        Args:
            cmd_name (CommandName_T): Command name to remove

        Returns:
            bool: Success or not
        """
        if cmd_name in cls._commands:
            cmd = cls._commands[cmd_name]
            for alias in list(
                filter(lambda x: cls._aliases[x] == cmd, cls._aliases.keys())
            ):
                del cls._aliases[alias]
            del cls._commands[cmd_name]
            if cmd in cls._switches:
                del cls._switches[cmd]
            return True
        return False

    @classmethod
    def switch_command_global(
        cls, cmd_name: CommandName_T, state: Optional[bool] = None
    ):
        """Change command state globally or simply switch it if `state` is None

        Args:
            cmd_name (CommandName_T): Command name
            state (Optional[bool]): State to change to. Defaults to None.
        """
        cmd = cls._commands[cmd_name]
        cls._switches[cmd] = not cls._switches[cmd] if state is None else bool(state)

    @classmethod
    def add_aliases(cls, aliases: Union[Iterable[str], str], cmd: Command):
        """Register command alias(es)

        Args:
            aliases (Union[Iterable[str], str]): Command aliases
            cmd (Command): Command
        """
        if isinstance(aliases, str):
            aliases = (aliases,)
        for alias in aliases:
            if not isinstance(alias, str):
                warnings.warn(f"Alias {alias} is not a string! Ignored")
                return
            elif alias in cls._aliases:
                warnings.warn(f"Alias {alias} already exists")
                return
            cls._aliases[alias] = cmd

    @classmethod
    def add_patterns(cls, patterns: Patterns_T, cmd: Command):
        """Register command alias(es)

        Args:
            patterns (Union[Iterable[Pattern], Pattern, Iterable[str], str]): Command patterns
            cmd (Command): Matched command
        """  # noqa: E501
        if isinstance(patterns, (str, Pattern)):
            patterns = (patterns,)
        for pattern in patterns:
            if isinstance(pattern, str):
                pattern = re.compile(pattern)
            if not isinstance(pattern, Pattern):
                warnings.warn(f"Pattern {pattern} is not a regex or string! Ignored")
                continue
            elif pattern in cls._patterns:
                warnings.warn(f"Pattern {pattern} already exists")
                continue
            cls._patterns[pattern] = cmd

    def _find_command(self, name: Union[str, CommandName_T]) -> Optional[Command]:
        cmd_name = (name,) if isinstance(name, str) else name
        if not cmd_name:
            return None

        cmd = {
            name: cmd
            for name, cmd in self.commands.items()
            if self.switches.get(cmd, True)
        }.get(cmd_name)
        return cmd

    def parse_command(
        self, bot: NoneBot, cmd_string: str
    ) -> Tuple[Optional[Command], Optional[str]]:
        logger.debug(f"Parsing command: {repr(cmd_string)}")

        matched_start = None
        for start in bot.config.COMMAND_START:
            # loop through COMMAND_START to find the longest matched start
            curr_matched_start = None
            if isinstance(start, str):
                if cmd_string.startswith(start):
                    curr_matched_start = start
            elif isinstance(start, Pattern):
                m = start.search(cmd_string)
                if m and m.start(0) == 0:
                    curr_matched_start = m.group(0)

            if curr_matched_start is not None and (
                matched_start is None or len(curr_matched_start) > len(matched_start)
            ):
                # a longer start, use it
                matched_start = curr_matched_start

        if matched_start is None:
            # it's not a command
            logger.debug("It's not a command")
            return None, None

        logger.debug(
            f"Matched command start: "
            f'{matched_start}{"(empty)" if not matched_start else ""}'
        )
        full_command = cmd_string[len(matched_start) :].lstrip()

        if not full_command:
            # command is empty
            return None, None

        cmd_name_text, *cmd_remained = full_command.split(maxsplit=1)

        cmd_name = None
        for sep in bot.config.COMMAND_SEP:
            # loop through COMMAND_SEP to find the most optimized split
            curr_cmd_name = None
            if isinstance(sep, str):
                curr_cmd_name = tuple(cmd_name_text.split(sep))
            elif isinstance(sep, Pattern):
                curr_cmd_name = tuple(sep.split(cmd_name_text))

            if curr_cmd_name is not None and (
                not cmd_name or len(curr_cmd_name) > len(cmd_name)
            ):
                # a more optimized split, use it
                cmd_name = curr_cmd_name

        if not cmd_name:
            cmd_name = (cmd_name_text,)
        logger.debug(f"Split command name: {cmd_name}")

        cmd = self._find_command(cmd_name)  # type: ignore
        if not cmd:
            logger.debug(f"Command {cmd_name} not found. Try to match aliases")
            cmd = {
                name: cmd
                for name, cmd in self.aliases.items()
                if self.switches.get(cmd, True)
            }.get(cmd_name_text)

        if not cmd:
            logger.debug(f"Alias {cmd_name} not found. Try to match patterns")
            patterns = {
                pattern: cmd
                for pattern, cmd in self.patterns.items()
                if self.switches.get(cmd, True)
            }
            for pattern in patterns:
                if pattern.search(full_command):
                    cmd = patterns[pattern]
                    logger.debug(
                        f"Pattern {pattern} of command {cmd.name} matched, "
                        f"function: {cmd.func}"
                    )
                    # if command matched by regex
                    # it will use the full_command as the current_arg of the session
                    return cmd, full_command

        if not cmd:
            return None, None

        logger.debug(f"Command {cmd.name} found, function: {cmd.func}")
        return cmd, "".join(cmd_remained)

    def switch_command(self, cmd_name: CommandName_T, state: Optional[bool] = None):
        """Change command state or simply switch it if `state` is None

        Args:
            cmd_name (CommandName_T): Command name
            state (Optional[bool]): State to change to. Defaults to None.
        """
        cmd = self.commands[cmd_name]
        self.switches[cmd] = not self.switches[cmd] if state is None else bool(state)


class CommandSession(BaseSession):
    __slots__ = (
        "cmd",
        "current_key",
        "current_arg_filters",
        "_current_send_kwargs",
        "current_arg",
        "_current_arg_text",
        "_current_arg_images",
        "_state",
        "_last_interaction",
        "_running",
        "_run_future",
        "_future",
    )

    def __init__(
        self,
        bot: NoneBot,
        event: CQEvent,
        cmd: Command,
        *,
        current_arg: Optional[str] = "",
        args: Optional[CommandArgs_T] = None,
    ):
        super().__init__(bot, event)
        self.cmd = cmd  # Command object

        # unique key of the argument that is currently requesting (asking)
        self.current_key: Optional[str] = None

        # initialize current argument filters
        self.current_arg_filters: Optional[List[Filter_T]] = None

        self._current_send_kwargs: Dict[str, Any] = {}

        # initialize current argument
        self.current_arg: Optional[str] = ""  # with potential CQ codes
        self._current_arg_text = None
        self._current_arg_images = None
        self.refresh(event, current_arg=current_arg)  # fill the above

        self._run_future = partial(asyncio.run_coroutine_threadsafe, loop=bot.loop)

        self._state: State_T = {}
        if args:
            self._state.update(args)

        self._last_interaction = None  # last interaction time of this session
        self._running = False
        self._future: Optional[asyncio.Future] = None

    @property
    def state(self) -> State_T:
        """
        State of the session.

        This contains all named arguments and
        other session scope temporary values.
        """
        return self._state

    @property
    def args(self) -> CommandArgs_T:
        """Deprecated. Use `session.state` instead."""
        return self.state

    @property
    def running(self) -> bool:
        """INTERNAL API"""
        return self._running

    @running.setter
    def running(self, value) -> None:
        """INTERNAL API"""
        if self._running is True and value is False:
            # change status from running to not running, record the time
            self._last_interaction = datetime.now()
        self._running = value

    @property
    def waiting(self) -> bool:
        """INTERNAL API"""
        return self._future is not None and not self._future.done()

    @property
    def expire_timeout(self) -> Optional[timedelta]:
        """INTERNAL API"""
        if self.cmd.expire_timeout is not ...:
            return self.cmd.expire_timeout
        return self.bot.config.SESSION_EXPIRE_TIMEOUT

    @property
    def run_timeout(self) -> Optional[timedelta]:
        """INTERNAL API"""
        if self.cmd.run_timeout is not ...:
            return self.cmd.run_timeout
        return self.bot.config.SESSION_RUN_TIMEOUT

    @property
    def is_valid(self) -> bool:
        """
        INTERNAL API

        Check whether the session has expired or not.
        """
        tm = self.expire_timeout
        if (
            tm
            and self._last_interaction
            and datetime.now() - self._last_interaction > tm
        ):
            return False
        return True

    @property
    def is_first_run(self) -> bool:
        return self._last_interaction is None

    @property
    def current_arg_text(self) -> str:
        """
        Plain text part in the current argument, without any CQ codes.
        """
        if self._current_arg_text is None:
            self._current_arg_text = Message(self.current_arg).extract_plain_text()
        return self._current_arg_text

    @property
    def current_arg_images(self) -> List[str]:
        """
        Images (as list of urls) in the current argument.
        """
        if self._current_arg_images is None:
            self._current_arg_images = [
                s.data["url"]
                for s in Message(self.current_arg)
                if s.type == "image" and "url" in s.data
            ]
        return self._current_arg_images

    @property
    def argv(self) -> List[str]:
        """
        Shell-like argument list, similar to sys.argv.
        Only available while shell_like is True in on_command decorator.
        """
        return self.state.get("argv", [])

    def refresh(self, event: CQEvent, *, current_arg: Optional[str] = "") -> None:
        """
        INTERNAL API

        Refill the session with a new message event.

        :param event: new message event
        :param current_arg: new command argument as a string
        """
        self.event = event
        self.current_arg = current_arg
        self._current_arg_text = None
        self._current_arg_images = None

    def get(
        self,
        key: str,
        *,
        prompt: Optional[Message_T] = None,
        arg_filters: Optional[List[Filter_T]] = None,
        **kwargs,
    ) -> Any:
        """
        Get an argument with a given key.

        If the argument does not exist in the current session,
        a pause exception will be raised, and the caller of
        the command will know it should keep the session for
        further interaction with the user.

        :param key: argument key
        :param prompt: prompt to ask the user
        :param arg_filters: argument filters for the next user input
        :param kwargs: other keyword arguments used in BeseSession.send()
        :return: the argument value
        """
        if key in self.state:
            return self.state[key]

        self.current_key = key
        self.current_arg_filters = arg_filters
        self._current_send_kwargs = kwargs
        self.pause(prompt, **kwargs)

    async def aget(
        self,
        key: str = ...,
        *,
        prompt: Optional[Message_T] = None,
        arg_filters: Optional[List[Filter_T]] = None,
        force_update: bool = ...,
        **kwargs,
    ) -> Any:
        """
        Get an argument with a given key.

        If the argument does not exist in the current session,
        the current coroutine yields, and the caller of
        the command will know it should keep the session for
        further interaction with the user.

        :param key: argument key
        :param prompt: prompt to ask the user
        :param arg_filters: argument filters for the next user input
        :param force_update: true to ignore the current argument
        :param kwargs: other keyword arguments used in BeseSession.send()
        :return: the argument value
        """
        if key is ...:
            key = "__default_argument"
            force_update = True
        if key in self.state:
            if force_update is True:
                del self.state[key]
            else:
                return self.state[key]

        self.current_key = key
        self.current_arg_filters = arg_filters
        self._current_send_kwargs = kwargs
        await self.apause(prompt, **kwargs)

        return self.state.get(key, self.current_arg)

    def get_optional(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        Simply get a argument with given key.

        Deprecated. Use `session.state.get()` instead.
        """
        return self.state.get(key, default)

    def pause(self, message: Optional[Message_T] = None, **kwargs) -> NoReturn:
        """
        Pause the session for further interaction. This function never returns.

        :param message: message to send to the user
        :param kwargs: other keyword arguments used in BeseSession.send()
        """
        if message:
            self._run_future(self.send(message, **kwargs))
        self._raise(_PauseException())

    async def apause(self, message: Optional[Message_T] = None, **kwargs) -> None:
        """
        Pause the session for further interaction. The control flow will pick
        up where it is left over when this command session is recalled.

        :param message: message to send to the user
        :param kwargs: other keyword arguments used in BeseSession.send()
        """
        if message:
            self._run_future(self.send(message, **kwargs))
        if self.waiting:
            return
        while True:
            try:
                self._future = asyncio.get_event_loop().create_future()
                self.running = False
                timeout_opt = self.expire_timeout
                timeout = timeout_opt.total_seconds() if timeout_opt else None
                await asyncio.wait_for(self._future, timeout)
                break
            except _PauseException:
                continue
            except (_FinishException, asyncio.TimeoutError):
                raise

    def finish(self, message: Optional[Message_T] = None, **kwargs) -> NoReturn:
        """
        Finish the session. This function never returns.

        :param message: message to send to the user when this command exits
        :param kwargs: other keyword arguments used in BeseSession.send()
        """
        if message:
            self._run_future(self.send(message, **kwargs))
        self._raise(_FinishException())

    def switch(self, new_message: Message_T) -> NoReturn:
        """
        Finish the session and switch to a new (fake) message event.

        The user may send another command (or another intention as natural
        language) when interacting with the current session. In this case,
        the session may not understand what the user is saying, so it
        should call this method and pass in that message, then NoneBot will
        handle the situation properly.
        """
        if self.is_first_run:
            # if calling this method during first run,
            # we think the command is not handled
            raise _FinishException(result=False)

        if not isinstance(new_message, Message):
            new_message = Message(new_message)
        self._raise(SwitchException(new_message))

    def _raise(self, e: Exception) -> NoReturn:
        """Raise an exception from the main execution path"""
        if self.waiting:
            self._future.set_exception(e)
            raise _YieldException
        raise e


async def handle_command(
    bot: NoneBot, event: CQEvent, manager: CommandManager
) -> Optional[bool]:
    """
    INTERNAL API

    Handle a message as a command.

    This function is typically called by "handle_message".

    :param bot: NoneBot instance
    :param event: message event
    :param manager: command manager
    :return: the message is handled as a command
    """
    cmd, current_arg = manager.parse_command(bot, str(event.message).lstrip())
    is_privileged_cmd = cmd and cmd.privileged
    if is_privileged_cmd and cmd.only_to_me and not event["to_me"]:
        is_privileged_cmd = False
    disable_interaction = bool(is_privileged_cmd)

    if is_privileged_cmd:
        logger.debug(f"Command {cmd.name} is a privileged command")

    ctx_id = context_id(event)

    if not is_privileged_cmd:
        # wait for 1.5 seconds (at most) if the current session is running
        retry = 5
        while retry > 0 and _sessions.get(ctx_id) and _sessions[ctx_id].running:
            retry -= 1
            await asyncio.sleep(0.3)

    check_perm = True
    session: Optional[CommandSession] = (
        _sessions.get(ctx_id) if not is_privileged_cmd else None
    )

    if session is not None:
        if session.running:
            logger.warning(
                f"There is a session of command "
                f"{session.cmd.name} running, notify the user"
            )
            asyncio.ensure_future(
                send(
                    bot, event, render_expression(bot.config.SESSION_RUNNING_EXPRESSION)
                )
            )
            # pretend we are successful, so that NLP won't handle it
            return True

        if session.is_valid:
            logger.debug(f"Session of command {session.cmd.name} exists")
            # since it's in a session, the user must be talking to me
            event["to_me"] = True
            session.refresh(event, current_arg=str(event["message"]))
            # there is no need to check permission for existing session
            check_perm = False
        else:
            # the session is expired, remove it
            logger.debug(f"Session of command {session.cmd.name} has expired")
            if ctx_id in _sessions:
                del _sessions[ctx_id]
            session = None

    if session is None:
        if not cmd:
            logger.debug("Not a known command, ignored")
            return False
        if cmd.only_to_me and not event["to_me"]:
            logger.debug("Not to me, ignored")
            return False
        SessionClass = cmd.session_class or CommandSession
        session = SessionClass(bot, event, cmd, current_arg=current_arg)
        logger.debug(f"New session of command {session.cmd.name} created")

    assert isinstance(session, CommandSession)
    return await _real_run_command(
        session, ctx_id, check_perm=check_perm, disable_interaction=disable_interaction
    )


async def call_command(
    bot: NoneBot,
    event: CQEvent,
    name: Union[str, CommandName_T],
    *,
    current_arg: str = "",
    args: Optional[CommandArgs_T] = None,
    check_perm: bool = True,
    disable_interaction: bool = False,
) -> Optional[bool]:
    """
    Call a command internally.

    This function is typically called by some other commands
    or "handle_natural_language" when handling NLPResult object.

    Note: If disable_interaction is not True, after calling this function,
    any previous command session will be overridden, even if the command
    being called here does not need further interaction (a.k.a asking
    the user for more info).

    :param bot: NoneBot instance
    :param event: message event
    :param name: command name
    :param current_arg: command current argument string
    :param args: command args
    :param check_perm: should check permission before running command
    :param disable_interaction: disable the command's further interaction
    :return: the command is successfully called
    """
    cmd = CommandManager()._find_command(name)
    if not cmd:
        return False
    SessionClass = cmd.session_class or CommandSession
    session = SessionClass(bot, event, cmd, current_arg=current_arg, args=args)
    return await _real_run_command(
        session,
        context_id(session.event),
        check_perm=check_perm,
        disable_interaction=disable_interaction,
    )


async def _real_run_command(
    session: CommandSession, ctx_id: str, disable_interaction: bool = False, **kwargs
) -> Optional[bool]:
    if not disable_interaction:
        # override session only when interaction is not disabled
        _sessions[ctx_id] = session
    try:
        logger.debug(f"Running command {session.cmd.name}")
        session.running = True
        future = asyncio.ensure_future(session.cmd.run(session, **kwargs))
        timeout_opt = session.run_timeout
        timeout = timeout_opt.total_seconds() if timeout_opt else None

        try:
            await asyncio.wait_for(future, timeout)
            handled = future.result()
        except asyncio.TimeoutError:
            handled = True
        except CommandInterrupt:
            raise
        except Exception as e:
            logger.error(
                f"An exception occurred while " f"running command {session.cmd.name}:"
            )
            logger.exception(e)
            handled = True
        raise _FinishException(handled)
    except _PauseException:
        session.running = False
        if disable_interaction:
            # if the command needs further interaction, we view it as failed
            return False
        logger.debug(f"Further interaction needed for " f"command {session.cmd.name}")
        # return True because this step of the session is successful
        return True
    except _YieldException:
        return True
        # return True because this step of the session is successful
    except (_FinishException, SwitchException) as e:
        session.running = False
        logger.debug(f"Session of command {session.cmd.name} finished")
        if not disable_interaction and ctx_id in _sessions:
            # the command is finished, remove the session,
            # but if interaction is disabled during this command call,
            # we leave the _sessions untouched.
            del _sessions[ctx_id]

        if isinstance(e, _FinishException):
            return e.result
        elif isinstance(e, SwitchException):
            # we are guaranteed that the session is not first run here,
            # which means interaction is definitely enabled,
            # so we can safely touch _sessions here.
            if ctx_id in _sessions:
                # make sure there is no session waiting
                del _sessions[ctx_id]
            logger.debug(
                f"Session of command {session.cmd.name} switching, "
                f"new message: {e.new_message}"
            )
            raise e  # this is intended to be propagated to handle_message()


def kill_current_session(event: CQEvent) -> None:
    """
    Force kill current session of the given event context,
    despite whether it is running or not.

    :param event: message event
    """
    ctx_id = context_id(event)
    if ctx_id in _sessions:
        if _sessions[ctx_id].waiting:
            _sessions[ctx_id]._future.set_exception(_FinishException())
        del _sessions[ctx_id]


from nonetrip.comp.command.group import CommandGroup  # noqa:E402

__all__ = [
    "CommandManager",
    "CommandSession",
    "call_command",
    "kill_current_session",
    "CommandGroup",
]



from typing import Optional, List


def _simple_chinese_to_bool(text: str) -> Optional[bool]:
    """
    Convert a chinese text to boolean.

    Examples:

        是的 -> True
        好的呀 -> True
        不要 -> False
        不用了 -> False
        你好呀 -> None
    """
    text = text.strip().lower().replace(' ', '') \
        .rstrip(',.!?~，。！？～了的呢吧呀啊呗啦')
    if text in {
            '要', '用', '是', '好', '对', '嗯', '行', 'ok', 'okay', 'yeah', 'yep',
            '当真', '当然', '必须', '可以', '肯定', '没错', '确定', '确认'
    }:
        return True
    if text in {
            '不', '不要', '不用', '不是', '否', '不好', '不对', '不行', '别', 'no', 'nono',
            'nonono', 'nope', '不ok', '不可以', '不能', '不可以'
    }:
        return False
    return None


def _split_nonempty_lines(text: str) -> List[str]:
    return list(filter(lambda x: x, text.splitlines()))


def _split_nonempty_stripped_lines(text: str) -> List[str]:
    return list(filter(lambda x: x, map(lambda x: x.strip(),
                                        text.splitlines())))


simple_chinese_to_bool = _simple_chinese_to_bool
split_nonempty_lines = _split_nonempty_lines
split_nonempty_stripped_lines = _split_nonempty_stripped_lines


__all__ = [
    'simple_chinese_to_bool',
    'split_nonempty_lines',
    'split_nonempty_stripped_lines',
]



from datetime import timedelta
from typing import Any, Dict, Tuple, Union, Callable, Iterable, Optional, Type

from nonetrip.comp.command import CommandSession
from nonetrip.comp.typing import CommandHandler_T, CommandName_T, Patterns_T

class CommandGroup:
    basename: Tuple[str]
    base_kwargs: Dict[str, Any]
    def __init__(
        self,
        name: Union[str, CommandName_T],
        *,
        permission: int = ...,
        only_to_me: bool = ...,
        privileged: bool = ...,
        shell_like: bool = ...,
        expire_timeout: Optional[timedelta] = ...,
        run_timeout: Optional[timedelta] = ...,
        session_class: Optional[Type[CommandSession]] = ...
    ): ...
    def command(
        self,
        name: Union[str, CommandName_T],
        *,
        aliases: Union[Iterable[str], str] = ...,
        patterns: Patterns_T = ...,
        permission: int = ...,
        only_to_me: bool = ...,
        privileged: bool = ...,
        shell_like: bool = ...,
        expire_timeout: Optional[timedelta] = ...,
        run_timeout: Optional[timedelta] = ...,
        session_class: Optional[Type[CommandSession]] = ...
    ) -> Callable[[CommandHandler_T], CommandHandler_T]: ...



from typing import Optional

from nonetrip.comp.typing import Message_T


class ValidateError(ValueError):
    def __init__(self, message: Optional[Message_T] = None):
        self.message = message


__all__ = [
    "ValidateError",
]



try:
    from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore
except ImportError:
    # APScheduler is not installed
    AsyncIOScheduler = None

if AsyncIOScheduler:

    class Scheduler(AsyncIOScheduler):
        pass


else:
    Scheduler = None  # type: ignore


__all__ = [
    "Scheduler",
]



import asyncio
import hashlib
import random
from typing import Any, Sequence

from nonebot.adapters.cqhttp.utils import escape

from ..poly import Event as CQEvent
from . import NoneBot
from .exceptions import CQHttpError
from .typing import Expression_T, Message_T


def context_id(event: CQEvent, *, mode: str = "default", use_hash: bool = False) -> str:
    """
    Calculate a unique id representing the context of the given event.

    mode:
      default: one id for one context
      group: one id for one group or discuss
      user: one id for one user

    :param event: the event object
    :param mode: unique id mode: "default", "group", or "user"
    :param use_hash: use md5 to hash the id or not
    """
    ctx_id = ""
    if mode == "default":
        if event.group_id:
            ctx_id = f"/group/{event.group_id}"
        elif event.discuss_id:
            ctx_id = f"/discuss/{event.discuss_id}"
        if event.user_id:
            ctx_id += f"/user/{event.user_id}"
    elif mode == "group":
        if event.group_id:
            ctx_id = f"/group/{event.group_id}"
        elif event.discuss_id:
            ctx_id = f"/discuss/{event.discuss_id}"
        elif event.user_id:
            ctx_id = f"/user/{event.user_id}"
    elif mode == "user":
        if event.user_id:
            ctx_id = f"/user/{event.user_id}"

    if ctx_id and use_hash:
        ctx_id = hashlib.md5(ctx_id.encode("ascii")).hexdigest()
    return ctx_id


async def send(
    bot: NoneBot,
    event: CQEvent,
    message: Message_T,
    *,
    ensure_private: bool = False,
    ignore_failure: bool = True,
    **kwargs,
) -> Any:
    """Send a message ignoring failure by default."""
    try:
        if ensure_private:
            event = event.copy()
            event["message_type"] = "private"
        return await bot.send(event, message, **kwargs)
    except CQHttpError:
        if not ignore_failure:
            raise
        return None


async def send_to_superusers(bot: NoneBot, message: Message_T, **kwargs) -> None:
    """
    Send a private message to all superusers that are defined in config.

    :param bot: nonebot object
    :param message: message to send to each superuser
    :param kwargs: keyword arguments used in bot.send_private_msg()
    """
    tasks = [
        bot.send_private_msg(user_id=user_id, message=message, **kwargs)
        for user_id in bot.config.SUPERUSERS
    ]
    await asyncio.gather(*tasks)


def render_expression(
    expr: Expression_T, *args, escape_args: bool = True, **kwargs
) -> str:
    """
    Render an expression to message string.

    :param expr: expression to render
    :param escape_args: should escape arguments or not
    :param args: positional arguments used in str.format()
    :param kwargs: keyword arguments used in str.format()
    :return: the rendered message
    """
    result: str
    if callable(expr):
        result = expr(*args, **kwargs)
    elif isinstance(expr, Sequence) and not isinstance(expr, str):
        result = random.choice(expr)
    else:
        result = expr  # type: ignore
    if escape_args:
        return result.format(
            *[escape(s) if isinstance(s, str) else s for s in args],
            **{k: escape(v) if isinstance(v, str) else v for k, v in kwargs.items()},
        )
    return result.format(*args, **kwargs)


__all__ = [
    "context_id",
    "send",
    "send_to_superusers",
    "render_expression",
]



from datetime import timedelta
from typing import Any, Dict, Iterable, Optional, Pattern, Set, Union

from nonebot import get_driver
from nonebot.config import Config
from pydantic import BaseModel, Extra

from nonetrip.comp.typing import Expression_T

NoneBotConfig: Config = get_driver().config


class DefaultConfig(BaseModel):
    class Config:
        extra = Extra.allow

    API_ROOT: str = ""
    ACCESS_TOKEN: str = ""
    SECRET: str = ""
    HOST: str = "127.0.0.1"
    PORT: int = 8080
    DEBUG: bool = True

    SUPERUSERS: Set[int] = set()
    NICKNAME: Union[str, Iterable[str]] = ""

    COMMAND_START: Iterable[Union[str, Pattern]] = {"/", "!", "／", "！"}
    COMMAND_SEP: Iterable[Union[str, Pattern]] = {"/", "."}

    SESSION_EXPIRE_TIMEOUT: Optional[timedelta] = timedelta(minutes=5)
    SESSION_RUN_TIMEOUT: Optional[timedelta] = None
    SESSION_RUNNING_EXPRESSION: Expression_T = "您有命令正在执行，请稍后再试"

    SHORT_MESSAGE_MAX_LENGTH: int = 50

    DEFAULT_VALIDATION_FAILURE_EXPRESSION: Expression_T = "您的输入不符合要求，请重新输入"
    MAX_VALIDATION_FAILURES: int = 3
    TOO_MANY_VALIDATION_FAILURES_EXPRESSION: Expression_T = "您输入错误太多次啦，如需重试，请重新触发本功能"

    SESSION_CANCEL_EXPRESSION: Expression_T = "好的"

    APSCHEDULER_CONFIG: Dict[str, Any] = {"apscheduler.timezone": "Asia/Shanghai"}



"""
Default configurations.

Any derived configurations must import everything from this module
at the very beginning of their code, and then set their own value
to override the default one.

For example:

>>> from nonetrip.comp.default_config import *
>>> PORT = 9090
>>> DEBUG = False
>>> SUPERUSERS.add(123456)
>>> NICKNAME = '小明'
"""

from datetime import timedelta
from typing import Collection, Union, Iterable, Pattern, Optional, Dict, Any

from .typing import Expression_T

API_ROOT: str = ""
ACCESS_TOKEN: str = ""
SECRET: str = ""
HOST: str = "127.0.0.1"
PORT: int = 8080
DEBUG: bool = True

SUPERUSERS: Collection[int] = set()
NICKNAME: Union[str, Iterable[str]] = ""

COMMAND_START: Iterable[Union[str, Pattern]] = {"/", "!", "／", "！"}
COMMAND_SEP: Iterable[Union[str, Pattern]] = {"/", "."}

SESSION_EXPIRE_TIMEOUT: Optional[timedelta] = timedelta(minutes=5)
SESSION_RUN_TIMEOUT: Optional[timedelta] = None
SESSION_RUNNING_EXPRESSION: Expression_T = "您有命令正在执行，请稍后再试"

SHORT_MESSAGE_MAX_LENGTH: int = 50

DEFAULT_VALIDATION_FAILURE_EXPRESSION: Expression_T = "您的输入不符合要求，请重新输入"
MAX_VALIDATION_FAILURES: int = 3
TOO_MANY_VALIDATION_FAILURES_EXPRESSION: Expression_T = "您输入错误太多次啦，如需重试，请重新触发本功能"

SESSION_CANCEL_EXPRESSION: Expression_T = "好的"

APSCHEDULER_CONFIG: Dict[str, Any] = {"apscheduler.timezone": "Asia/Shanghai"}


__all__ = [
    "API_ROOT",
    "ACCESS_TOKEN",
    "SECRET",
    "HOST",
    "PORT",
    "DEBUG",
    "SUPERUSERS",
    "NICKNAME",
    "COMMAND_START",
    "COMMAND_SEP",
    "SESSION_EXPIRE_TIMEOUT",
    "SESSION_RUN_TIMEOUT",
    "SESSION_RUNNING_EXPRESSION",
    "SHORT_MESSAGE_MAX_LENGTH",
    "DEFAULT_VALIDATION_FAILURE_EXPRESSION",
    "MAX_VALIDATION_FAILURES",
    "TOO_MANY_VALIDATION_FAILURES_EXPRESSION",
    "SESSION_CANCEL_EXPRESSION",
    "APSCHEDULER_CONFIG",
]



from argparse import ArgumentParser as BaseArgumentParser

from .command import CommandSession


class ParserExit(RuntimeError):
    """INTERNAL API"""

    def __init__(self, status=0, message=None):
        self.status = status
        self.message = message


class ArgumentParser(BaseArgumentParser):
    """
    An ArgumentParser wrapper that avoid printing messages to
    standard I/O.
    """

    def __init__(self, *args, **kwargs):
        self.session = kwargs.pop("session", None)
        super().__init__(*args, **kwargs)

    def _session_finish(self, message):
        if self.session and isinstance(self.session, CommandSession):
            self.session.finish(message)

    def _print_message(self, message, file=None):
        # do nothing
        pass

    def exit(self, status=0, message=None):
        raise ParserExit(status=status, message=message)

    def parse_args(self, args=None, namespace=None):
        try:
            return super().parse_args(args=args, namespace=namespace)
        except ParserExit as e:
            if e.status == 0:
                # --help
                self._session_finish(self.usage or self.format_help())
            else:
                self._session_finish("参数不足或不正确，请使用 --help 参数查询使用帮助")


__all__ = [
    "ArgumentParser",
]



from typing import Union, Callable

from nonetrip.comp.plugin import on_command
from nonetrip.comp.typing import CommandHandler_T, CommandName_T


class CommandGroup:
    """
    Group a set of commands with same name prefix.
    """

    __slots__ = ("basename", "base_kwargs")

    def __init__(self, name: Union[str, CommandName_T], **kwargs):
        if "aliases" in kwargs or "patterns" in kwargs:
            raise ValueError(
                "aliases or patterns should not be used as base kwargs for group"
            )
        self.basename = (name,) if isinstance(name, str) else name
        self.base_kwargs = kwargs

    def command(
        self, name: Union[str, CommandName_T], **kwargs
    ) -> Callable[[CommandHandler_T], CommandHandler_T]:
        """
        Decorator to register a function as a command. Its has the same usage as
        `on_command`.

        :param kwargs: keyword arguments will be passed to `on_command`. For each
                       argument in the signature of this method here, if it is not
                       present when calling, default value for the command group is
                       used (e.g. `self.permission`). If that value is also not set,
                       default value for `on_command` is used.
        """
        sub_name = (name,) if isinstance(name, str) else name
        name = self.basename + sub_name

        final_kwargs = {**self.base_kwargs, **kwargs}
        return on_command(name, **final_kwargs)



"""
Provide logger object.

Any other modules in "nonebot" should use "logger" from this module
to log messages.
"""

import logging

from nonebot.log import LoguruHandler

logger = logging.getLogger("nonebot")
default_handler = LoguruHandler()
default_handler.setFormatter(
    logging.Formatter("[%(asctime)s %(name)s] %(levelname)s: %(message)s")
)
logger.addHandler(default_handler)


__all__ = [
    "logger",
]



"""
This module replicates the standard permission module already defined
in nonetrip.comp. This, yet another implementation, is experimental and may
be easier or harder to use than the standard one.
"""

import asyncio
from datetime import datetime, time
from typing import (
    Any,
    Awaitable,
    Callable,
    Container,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Union,
)

from aiocache.decorators import cached
from nonetrip.comp import NoneBot
from nonetrip.comp.exceptions import CQHttpError

from ...poly import Event as CQEvent


class SenderRoles(NamedTuple):
    """
    A high level agent object to assess a message sender's permissions
    such as message type, sender's group role, etc. If any groups are
    involved, such as group messages and direct private messages from
    a group member, that group number is consistent across all methods
    """

    bot: NoneBot
    event: CQEvent
    sender: Optional[Dict[str, Any]]

    @staticmethod
    async def create(bot: NoneBot, event: CQEvent) -> "SenderRoles":
        """constructor to create a SenderRoles object from an event"""
        # same approach as vanilla permission checker
        assert event.group_id and event.user_id
        sender_info = (
            await _get_member_info(bot, event.self_id, event.group_id, event.user_id)
            if event.get("message_type") == "group"
            else None
        )
        return SenderRoles(bot, event, sender_info)

    # builtin components:

    @property
    def is_superuser(self) -> bool:
        return self.event.user_id in self.bot.config.SUPERUSERS

    @property
    def is_groupchat(self) -> bool:
        return self.event.get("message_type") == "group"

    @property
    def is_anonymous(self) -> bool:
        return self.event.sub_type == "anonymous"

    @property
    def is_admin(self) -> bool:
        return self.sender is not None and self.sender.get("role") == "admin"

    @property
    def is_owner(self) -> bool:
        return self.sender is not None and self.sender.get("role") == "owner"

    @property
    def is_privatechat(self) -> bool:
        return self.event.get("message_type") == "private"

    @property
    def is_private_friend(self) -> bool:
        return self.is_privatechat and self.event.sub_type == "friend"

    @property
    def is_private_group(self) -> bool:
        return self.is_privatechat and self.event.sub_type == "group"

    @property
    def is_private_discuss(self) -> bool:
        return self.is_privatechat and self.event.sub_type == "discuss"

    @property
    def is_discusschat(self) -> bool:
        return self.event.get("message_type") == "discuss"

    def from_group(self, group_id: Union[int, Container[int]]) -> bool:
        """returns True if the sender belongs to these groups (group_ids)"""
        if isinstance(group_id, int):
            return self.event.group_id == group_id
        return self.event.group_id in group_id

    def sent_by(self, sender_id: Union[int, Container[int]]) -> bool:
        """returns True if the sender is one of these people (sender_ids)"""
        if isinstance(sender_id, int):
            return self.event.user_id == sender_id
        return self.event.user_id in sender_id


# AS OF this commit: same as original _get_member_info()
@cached(ttl=2 * 60)
async def _get_member_info(
    bot: NoneBot, self_id: int, group_id: int, user_id: int
) -> Optional[Dict[str, Any]]:
    try:
        return await bot.get_group_member_info(
            self_id=self_id, group_id=group_id, user_id=user_id, no_cache=True
        )
    except CQHttpError:
        return None


async def check_permission(
    bot: NoneBot, event: CQEvent, policy: "RoleCheckPolicy"
) -> bool:
    """
    Check whether the message sender has permission required defined
    by the policy

    :param bot: NoneBot instance
    :param event: message event
    :param policy: the policy that returns boolean
    :return: the context has the permission
    """
    sender_roles = await SenderRoles.create(bot, event)
    res = policy(sender_roles)
    if isinstance(res, Awaitable):
        return await res
    return res


# Polices are the functions that triggers when events arrive. They return
# booleans to determine whether to proceed with the commands. Here we give
# its definition, and some example implementations.


RoleCheckPolicy = Callable[[SenderRoles], Union[bool, Awaitable[bool]]]


def aggregate_policy(
    policies: Iterable[RoleCheckPolicy],
    aggregator: Callable[[Iterable[object]], bool] = all,
) -> RoleCheckPolicy:
    """
    Merge several role checkers into one using the AND operator (if `aggregator`
    is builtin function `all` - by default). if all given policies are sync,
    the merged one is sync, otherwise it is async. This is useful to concatenate
    policies like applying [blocklist, groupchat, ...] altogether.

    async functions are slow maybe. use with caution.

    The `aggregator` parameter is recommended to be the bultin functions such as
    `all` (performing AND operations on each checkers), or `any` (performing OR
    operations) because they short circuit. However it is possible to define more
    complex aggregators.

    Be aware that the order of checkers being called is not specified.

    :param policies: list of policies
    :param aggregator: the function used to combine the results of each separate
                       checkers as items consumed in iterators.
    :return: new policy
    """
    syncs = []  # type: List[Callable[[SenderRoles], bool]]
    asyncs = []  # type: List[Callable[[SenderRoles], Awaitable[bool]]]
    for f in policies:
        if asyncio.iscoroutinefunction(f) or (
            asyncio.iscoroutinefunction(getattr(f, "__call__", None))
        ):
            # pyright cannot narrow down types
            asyncs.append(f)  # type: ignore
        else:
            syncs.append(f)  # type: ignore

    def checker_sync(sender: SenderRoles) -> bool:
        return aggregator(f(sender) for f in syncs)

    if len(asyncs) == 0:
        return checker_sync

    async def checker_async(sender: SenderRoles) -> bool:
        if not checker_sync(sender):
            return False
        # no short circuiting currently :-(
        coros = [f(sender) for f in asyncs]
        return aggregator(await asyncio.gather(*coros))

    return checker_async


def simple_allow_list(
    *,
    user_ids: Container[int] = ...,  # type:ignore
    group_ids: Container[int] = ...,  # type: ignore
    reverse: bool = False
) -> RoleCheckPolicy:
    """
    Creates a policy that only allows senders from these users or groups.
    The returned function is stateless.

    ctrl+mouse to this function to see how to write a complex policy.

    :param user_ids: set of user ids to allow
    :param group_ids: set of group ids to allow
    :param reverse: if this is true, then bans the aforementioned
                    senders instead (policy becomes blocklist)
    :return: new policy
    """
    user_ids = user_ids if user_ids is not ... else set()
    group_ids = group_ids if group_ids is not ... else set()

    def checker(sender: SenderRoles) -> bool:
        is_in = sender.sent_by(user_ids) or sender.from_group(group_ids)
        return not is_in if reverse else is_in

    return checker


def simple_time_range(
    begin_time: time, end_time: time, reverse: bool = False, tz_info: Any = None
) -> RoleCheckPolicy:
    """
    Creates a policy that only allows commands to be activated between
    begin_time and end_time. Uses builtin datetime.datetime module.
    The returned function is stateless.

    :param begin_time: starting time and
    :param end_time: ending time to allow users
    :param reverse: if this is true, then bans the aforementioned
                    time ranges instead
    :param tz_info: argument to pass to datetime.now()
    :return: new policy
    """
    # source: https://stackoverflow.com/questions/10048249
    def checker(_: Any) -> bool:
        now = datetime.now(tz_info).time()
        if begin_time < end_time:
            is_in = now >= begin_time and now <= end_time
        else:
            is_in = now >= begin_time or now <= end_time
        return not is_in if reverse else is_in

    return checker


__all__ = [
    "SenderRoles",
    "check_permission",
    "RoleCheckPolicy",
    "aggregate_policy",
    "simple_allow_list",
    "simple_time_range",
]



import re

from nonetrip.comp import CommandSession
from nonetrip.comp.helpers import render_expression


def handle_cancellation(session: CommandSession):
    """
    If the input is a string of cancellation word, finish the command session.
    """

    def control(value):
        if _is_cancellation(value) is True:
            session.finish(
                render_expression(session.bot.config.SESSION_CANCEL_EXPRESSION)
            )
        return value

    return control


def _is_cancellation(sentence: str) -> bool:
    for kw in ("算", "别", "不", "停", "取消"):
        if kw in sentence:
            # a keyword matches
            break
    else:
        # no keyword matches
        return False

    if re.match(r"^那?[算别不停]\w{0,3}了?吧?$", sentence) or re.match(
        r"^那?(?:[给帮]我)?取消了?吧?$", sentence
    ):
        return True

    return False


__all__ = [
    "handle_cancellation",
]



from nonetrip.comp import CommandSession, on_command
from nonetrip.comp import permission as perm
from nonetrip.comp.message import unescape


@on_command("echo")
async def echo(session: CommandSession):
    await session.send(session.state.get("message") or session.current_arg)


@on_command("say", permission=perm.SUPERUSER)
async def say(session: CommandSession):
    await session.send(unescape(session.state.get("message") or session.current_arg))



# flake8:noqa:E402
import asyncio
import logging
from typing import Any, Awaitable, Callable, Optional

from nonebot import get_driver
from nonebot.adapters.cqhttp.message import Message

from ..config import DefaultConfig
from ..poly import CQHttp, Event
from . import default_config
from .log import logger
from .sched import Scheduler

if Scheduler:
    scheduler = Scheduler()
else:
    scheduler = None


class NoneBot(CQHttp):
    def __init__(self, config_object: Optional[Any] = None):
        super().__init__()

        config_object = config_object or default_config
        config_dict = {
            k: getattr(config_object, k)
            for k in dir(config_object)
            if k.isupper() and not k.startswith("_")
        }

        self.config = DefaultConfig.parse_obj(config_dict)
        logger.debug(f"Loaded configurations: {config_dict}")

        self.logger.setLevel(logging.DEBUG if self.config.DEBUG else logging.INFO)

        from .message import handle_message
        from .notice_request import handle_notice_or_request

        @self.on_message
        async def _msg_handler(selfevent: Event):
            asyncio.create_task(handle_message(self, selfevent))

        @self.on_notice
        async def _not_handler(selfevent: Event):
            asyncio.create_task(handle_notice_or_request(self, selfevent))

        @self.on_request
        async def _req_handler(selfevent: Event):
            asyncio.create_task(handle_notice_or_request(self, selfevent))


_bot: Optional[NoneBot] = None


def init(config: Optional[Any] = None, enable_scheduler: bool = True):
    global _bot
    _bot = NoneBot(config)

    if enable_scheduler and scheduler and not scheduler.running:
        scheduler.configure(_bot.config.APSCHEDULER_CONFIG)
        scheduler.start()
        logger.info("Scheduler started")

    return _bot


def get_bot() -> NoneBot:
    """
    Get the NoneBot instance.

    The result is ensured to be not None, otherwise an exception will
    be raised.

    :raise ValueError: instance not initialized
    """
    if _bot is None:
        raise ValueError("NoneBot instance has not been initialized")
    return _bot


def on_startup(func: Callable[[], Awaitable[None]]) -> Callable[[], Awaitable[None]]:
    """
    Decorator to register a function as startup callback.
    """
    return get_driver().on_startup(func)


def on_websocket_connect(
    func: Callable[[Event], Awaitable[None]]
) -> Callable[[], Awaitable[None]]:
    """
    Decorator to register a function as websocket connect callback.

    Only work with CQHTTP v4.14+.
    """
    return get_bot().on_meta_event("lifecycle.connect")(func)  # type: ignore


from .command import CommandGroup, CommandSession
from .exceptions import CQHttpError
from .helpers import context_id
from .message import Message, MessageSegment, message_preprocessor
from .natural_language import IntentCommand, NLPResult, NLPSession
from .notice_request import NoticeSession, RequestSession
from .plugin import (get_loaded_plugins, load_builtin_plugins, load_plugin,
                     load_plugins, on_command, on_natural_language, on_notice,
                     on_request)

__all__ = [
    "NoneBot",
    "scheduler",
    "init",
    "get_bot",
    "on_startup",
    "on_websocket_connect",
    "CQHttpError",
    "load_plugin",
    "load_plugins",
    "load_builtin_plugins",
    "get_loaded_plugins",
    "message_preprocessor",
    "Message",
    "MessageSegment",
    "on_command",
    "CommandSession",
    "CommandGroup",
    "on_natural_language",
    "NLPSession",
    "NLPResult",
    "IntentCommand",
    "on_notice",
    "NoticeSession",
    "on_request",
    "RequestSession",
    "context_id",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonetrip',
 'nonetrip.comp',
 'nonetrip.comp.command',
 'nonetrip.comp.command.argfilter',
 'nonetrip.comp.experimental',
 'nonetrip.comp.plugins']

package_data = \
{'': ['*']}

install_requires = \
['aiocache>=0.11.1,<0.12.0', 'nonebot2>=2.0.0-alpha.10,<3.0.0']

extras_require = \
{'scheduler': ['APScheduler>=3.7.0,<4.0.0']}

setup_kwargs = {
    'name': 'nonetrip',
    'version': '0.1.0',
    'description': 'A compatibility layer plug-in on NoneBot2 which can provide compatibility with NoneBot1',
    'long_description': None,
    'author': 'mixmoe',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
)

from ..poly import Event as CQEvent

if TYPE_CHECKING:
    from nonetrip.comp import NoneBot
    from nonetrip.comp.command import CommandSession
    from nonetrip.comp.natural_language import IntentCommand, NLPSession
    from nonetrip.comp.notice_request import NoticeSession, RequestSession
    from nonetrip.comp.plugin import PluginManager


Context_T = Dict[str, Any]
Message_T = Union[str, Dict[str, Any], List[Dict[str, Any]]]
Expression_T = Union[str, Sequence[str], Callable[..., str]]
CommandName_T = Tuple[str, ...]
CommandArgs_T = Dict[str, Any]
CommandHandler_T = Callable[["CommandSession"], Awaitable[Any]]
Patterns_T = Union[Iterable[str], str, Iterable[Pattern[str]], Pattern[str]]
State_T = Dict[str, Any]
Filter_T = Callable[[Any], Union[Any, Awaitable[Any]]]
PermChecker_T = Callable[["NoneBot", "CQEvent"], Awaitable[bool]]
NLPHandler_T = Callable[["NLPSession"], Awaitable[Optional["IntentCommand"]]]
RequestHandler_T = Callable[["RequestSession"], Awaitable[Any]]
MessagePreprocessor_T = Callable[
    ["NoneBot", "CQEvent", "PluginManager"], Awaitable[Any]
]
MessagePreprocessor_T = Callable[
    ["NoneBot", "CQEvent", "PluginManager"], Awaitable[Any]
]
NoticeHandler_T = Callable[["NoticeSession"], Awaitable[Any]]

__all__ = [
    "Context_T",
    "Message_T",
    "Expression_T",
    "CommandName_T",
    "CommandArgs_T",
    "CommandHandler_T",
    "Patterns_T",
    "State_T",
    "Filter_T",
    "PermChecker_T",
    "PermChecker_T",
]



from typing import List, Union

from ..bus import EventBus
from ..poly import Event as CQEvent
from . import NoneBot
from .exceptions import CQHttpError
from .log import logger
from .session import BaseSession
from .typing import NoticeHandler_T, RequestHandler_T

_bus = EventBus()


class EventHandler:
    """INTERNAL API"""

    __slots__ = ("events", "func")

    def __init__(
        self, events: List[str], func: Union[NoticeHandler_T, RequestHandler_T]
    ):
        self.events = events
        self.func = func


class NoticeSession(BaseSession):
    __slots__ = ()

    def __init__(self, bot: NoneBot, event: CQEvent):
        super().__init__(bot, event)


class RequestSession(BaseSession):
    __slots__ = ()

    def __init__(self, bot: NoneBot, event: CQEvent):
        super().__init__(bot, event)

    async def approve(self, remark: str = "") -> None:
        """
        Approve the request.

        :param remark: remark of friend (only works in friend request)
        """
        try:
            await self.bot.call_action(
                action=".handle_quick_operation_async",
                self_id=self.event.self_id,
                context=self.event,
                operation={"approve": True, "remark": remark},
            )
        except CQHttpError:
            pass

    async def reject(self, reason: str = "") -> None:
        """
        Reject the request.

        :param reason: reason to reject (only works in group request)
        """
        try:
            await self.bot.call_action(
                action=".handle_quick_operation_async",
                self_id=self.event.self_id,
                context=self.event,
                operation={"approve": False, "reason": reason},
            )
        except CQHttpError:
            pass


async def handle_notice_or_request(bot: NoneBot, event: CQEvent) -> None:
    """INTERNAL API"""
    if event.type == "notice":
        _log_notice(event)
        session = NoticeSession(bot, event)
    else:  # must be 'request'
        _log_request(event)
        session = RequestSession(bot, event)

    ev_name = event.name
    logger.debug(f"Emitting event: {ev_name}")
    try:
        await _bus.emit(ev_name, session)
    except Exception as e:
        logger.error(f"An exception occurred while handling event {ev_name}:")
        logger.exception(e)


def _log_notice(event: CQEvent) -> None:
    logger.info(f"Notice: {event}")


def _log_request(event: CQEvent) -> None:
    logger.info(f"Request: {event}")


__all__ = [
    "NoticeSession",
    "RequestSession",
]
